"""
Copyright 2019 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import base64
import functools
import json
import logging
import os
import uuid
from datetime import UTC, datetime, timedelta, timezone
from functools import partial

import pytest
from dateutil import parser
from tornado.httpclient import AsyncHTTPClient, HTTPRequest

from inmanta import config, const, data, loader, resources, util
from inmanta.agent import executor, handler
from inmanta.const import ParameterSource
from inmanta.data import AUTO_DEPLOY, ResourcePersistentState
from inmanta.data.model import AttributeStateChange
from inmanta.deploy import persistence, state
from inmanta.export import upload_code
from inmanta.protocol import Client
from inmanta.server import SLICE_AGENT_MANAGER, SLICE_ORCHESTRATION, SLICE_SERVER
from inmanta.server import config as opt
from inmanta.server.bootloader import InmantaBootloader
from inmanta.types import ResourceIdStr, ResourceVersionIdStr
from inmanta.util import get_compiler_version
from utils import log_contains, log_doesnt_contain, retry_limited

LOGGER = logging.getLogger(__name__)


@pytest.mark.parametrize("no_agent", [True])
@pytest.mark.parametrize(
    "n_versions_to_keep, n_versions_to_create",
    [
        (2, 4),
        (4, 2),
        (2, 2),
    ],
)
async def test_create_too_many_versions(client, server, no_agent, n_versions_to_keep, n_versions_to_create):
    """
    - set AVAILABLE_VERSIONS_TO_KEEP environment setting to <n_versions_to_keep>
    - create <n_versions_to_create> versions
    - check the actual number of versions before and after cleanup
    """

    # Create project
    result = await client.create_project("env-test")
    assert result.code == 200
    project_id = result.result["project"]["id"]

    # Create environment
    result = await client.create_environment(project_id=project_id, name="env_1")
    env_1_id = result.result["environment"]["id"]
    result = await client.set_setting(tid=env_1_id, id=data.AVAILABLE_VERSIONS_TO_KEEP, value=n_versions_to_keep)
    assert result.code == 200
    # Perform release manually using the release_version() endpoint to prevent race conditions.
    result = await client.set_setting(env_1_id, AUTO_DEPLOY, False)
    assert result.code == 200

    # make a second environment to be sure we don't do cross env deletes
    # as it is empty, if it leaks, it will likely take everything with it on the other one
    await client.create_environment(project_id=project_id, name="env_2")

    # Check value was set
    result = await client.get_setting(tid=env_1_id, id=data.AVAILABLE_VERSIONS_TO_KEEP)
    assert result.code == 200
    assert result.result["value"] == n_versions_to_keep

    for _ in range(n_versions_to_create):
        version = (await client.reserve_version(env_1_id)).result["data"]

        resources = [
            # First one is fixed
            {
                "id": f"std::testing::NullResource[vm1.dev.inmanta.com,name=network],v={version}",
                "owner": "root",
                "path": "/etc/sysconfig/network",
                "permissions": 644,
                "purged": False,
                "requires": [],
            },
            # This one changes ID every version
            {
                "id": f"std::testing::NullResource[vm1.dev.inmanta.com,name=network{version}],v={version}",
                "owner": "root",
                "path": "/etc/sysconfig/network",
                "permissions": 644,
                "purged": False,
                "requires": [],
            },
        ]

        res = await client.put_version(
            tid=env_1_id,
            version=version,
            resources=resources,
            unknowns=[],
            version_info={},
            compiler_version=get_compiler_version(),
        )
        assert res.code == 200

        res = await client.release_version(tid=env_1_id, id=version)
        assert res.code == 200

    versions = await client.list_versions(tid=env_1_id)
    assert versions.result["count"] == n_versions_to_create

    prvs = await ResourcePersistentState.get_list()
    assert len(prvs) == n_versions_to_create + 1

    # Ensure we don't clean too much
    await ResourcePersistentState.trim(env_1_id)

    prvs = await ResourcePersistentState.get_list()
    assert len(prvs) == n_versions_to_create + 1

    await server.get_slice(SLICE_ORCHESTRATION)._purge_versions()

    versions = await client.list_versions(tid=env_1_id)
    # Add +1, because the latest released version is not included in the AVAILABLE_VERSIONS_TO_KEEP setting.
    assert versions.result["count"] == min(n_versions_to_keep + 1, n_versions_to_create)

    prvs = await ResourcePersistentState.get_list()
    # 1 variable resource per version + 1 fixed resource per version
    assert len(prvs) == min(n_versions_to_keep + 2, n_versions_to_create + 1)


@pytest.mark.parametrize("has_released_versions", [True, False])
async def test_purge_versions(server, client, environment, has_released_versions: bool, agent_no_state_check) -> None:
    """
    Verify that the `OrchestrationService._purge_versions()` method works correctly and that it doesn't cleanup
    the latest released version.
    """
    result = await client.set_setting(tid=environment, id=data.AUTO_DEPLOY, value="false")
    assert result.code == 200

    versions = []
    for _ in range(5):
        version = (await client.reserve_version(environment)).result["data"]
        versions.append(version)
        res = await client.put_version(
            tid=environment,
            version=version,
            resources=[
                {
                    "id": f"unittest::Resource[internal,name=ok],v={version}",
                    "name": "root",
                    "desired_value": "ok",
                    "send_event": "false",
                    "purged": False,
                    "requires": [],
                }
            ],
            unknowns=[],
            version_info={},
            compiler_version=get_compiler_version(),
        )
        assert res.code == 200

    if has_released_versions:
        for v in versions[0:2]:
            result = await client.release_version(environment, id=v)
            assert result.code == 200

    result = await client.set_setting(tid=environment, id=data.AVAILABLE_VERSIONS_TO_KEEP, value=3)
    assert result.code == 200
    await server.get_slice(SLICE_ORCHESTRATION)._purge_versions()

    result = await client.list_versions(environment)
    assert result.code == 200
    assert result.result["count"] == (4 if has_released_versions else 3)
    if has_released_versions:
        assert {v["version"] for v in result.result["versions"]} == {versions[1], *versions[2:]}
    else:
        assert {v["version"] for v in result.result["versions"]} == {*versions[2:]}

    result = await client.set_setting(tid=environment, id=data.AVAILABLE_VERSIONS_TO_KEEP, value=1)
    assert result.code == 200
    await server.get_slice(SLICE_ORCHESTRATION)._purge_versions()

    result = await client.list_versions(environment)
    assert result.code == 200
    assert result.result["count"] == (2 if has_released_versions else 1)
    if has_released_versions:
        assert {v["version"] for v in result.result["versions"]} == {versions[1], *versions[4:]}
    else:
        assert {v["version"] for v in result.result["versions"]} == {*versions[4:]}


async def test_n_versions_env_setting_scope(client, server):
    """
    The AVAILABLE_VERSIONS_TO_KEEP environment setting used to be a global config option.
    This test checks that a specific environment setting can be set for each environment
    """

    n_versions_to_keep_env1 = 5
    n_versions_to_keep_env2 = 2

    n_many_versions = n_versions_to_keep_env1 + n_versions_to_keep_env2

    # Create project
    result = await client.create_project("env-test")
    assert result.code == 200
    project_id = result.result["project"]["id"]

    # Create environments
    result = await client.create_environment(project_id=project_id, name="env_1")
    env_1_id = result.result["environment"]["id"]
    result = await client.set_setting(tid=env_1_id, id=data.AVAILABLE_VERSIONS_TO_KEEP, value=n_versions_to_keep_env1)
    assert result.code == 200
    # Make sure we don't have a released version. _purge_versions() always keeps the latest released version.
    result = await client.set_setting(env_1_id, AUTO_DEPLOY, False)
    assert result.code == 200

    result = await client.create_environment(project_id=project_id, name="env_2")
    env_2_id = result.result["environment"]["id"]
    result = await client.set_setting(tid=env_2_id, id=data.AVAILABLE_VERSIONS_TO_KEEP, value=n_versions_to_keep_env2)
    assert result.code == 200
    # Make sure we don't have a released version. _purge_versions() always keeps the latest released version.
    result = await client.set_setting(env_2_id, AUTO_DEPLOY, False)
    assert result.code == 200

    # Create a lot of versions in both environments
    for _ in range(n_many_versions):
        env1_version = (await client.reserve_version(env_1_id)).result["data"]
        env2_version = (await client.reserve_version(env_2_id)).result["data"]

        res = await client.put_version(
            tid=env_1_id,
            version=env1_version,
            resources=[],
            unknowns=[],
            version_info={},
            compiler_version=get_compiler_version(),
        )
        assert res.code == 200

        res = await client.put_version(
            tid=env_2_id,
            version=env2_version,
            resources=[],
            unknowns=[],
            version_info={},
            compiler_version=get_compiler_version(),
        )
        assert res.code == 200

    # Before cleanup we have too many versions in both envs
    versions = await client.list_versions(tid=env_1_id)
    assert versions.result["count"] == n_many_versions

    versions = await client.list_versions(tid=env_2_id)
    assert versions.result["count"] == n_many_versions

    # Cleanup
    await server.get_slice(SLICE_ORCHESTRATION)._purge_versions()

    # After cleanup each env should have its specific number of version
    versions = await client.list_versions(tid=env_1_id)
    assert versions.result["count"] == n_versions_to_keep_env1

    versions = await client.list_versions(tid=env_2_id)
    assert versions.result["count"] == n_versions_to_keep_env2


@pytest.mark.slowtest
async def test_resource_action_update(server_multi, client_multi, environment_multi, null_agent_multi):
    """
    Test the server to manage the updates on a model during agent deploy
    """
    aclient = null_agent_multi._client
    version = (await client_multi.reserve_version(environment_multi)).result["data"]

    resources = [
        {
            "group": "root",
            "hash": "89bf880a0dc5ffc1156c8d958b4960971370ee6a",
            "id": "std::testing::NullResource[vm1.dev.inmanta.com,name=network],v=%d" % version,
            "owner": "root",
            "path": "/etc/sysconfig/network",
            "permissions": 644,
            "purged": False,
            "reload": False,
            "requires": [],
            "version": version,
        },
        {
            "group": "root",
            "hash": "b4350bef50c3ec3ee532d4a3f9d6daedec3d2aba",
            "id": "std::testing::NullResource[vm2.dev.inmanta.com,name=motd],v=%d" % version,
            "owner": "root",
            "path": "/etc/motd",
            "permissions": 644,
            "purged": False,
            "reload": False,
            "requires": [],
            "version": version,
        },
        {
            "group": "root",
            "hash": "3bfcdad9ab7f9d916a954f1a96b28d31d95593e4",
            "id": "std::testing::NullResource[vm1.dev.inmanta.com,name=hostname],v=%d" % version,
            "owner": "root",
            "path": "/etc/hostname",
            "permissions": 644,
            "purged": False,
            "reload": False,
            "requires": [],
            "version": version,
        },
        {
            "id": "std::Service[vm1.dev.inmanta.com,name=network],v=%d" % version,
            "name": "network",
            "onboot": True,
            "requires": ["std::testing::NullResource[vm1.dev.inmanta.com,name=network],v=%d" % version],
            "state": "running",
            "version": version,
        },
    ]

    res = await client_multi.put_version(
        tid=environment_multi,
        version=version,
        resources=resources,
        unknowns=[],
        version_info={},
        compiler_version=get_compiler_version(),
    )
    assert res.code == 200

    result = await client_multi.list_versions(environment_multi)
    assert result.code == 200
    assert result.result["count"] == 1

    result = await client_multi.release_version(environment_multi, version, False)
    assert result.code == 200

    result = await client_multi.get_version(environment_multi, version)
    assert result.code == 200
    assert result.result["model"]["version"] == version
    assert result.result["model"]["total"] == len(resources)
    assert result.result["model"]["released"]

    action_id = uuid.uuid4()
    now = datetime.now()
    result = await aclient.resource_action_update(
        environment_multi,
        ["std::testing::NullResource[vm1.dev.inmanta.com,name=network],v=%d" % version],
        action_id,
        "deploy",
        now,
        now,
        "deployed",
        [],
        {},
    )

    assert result.code == 200

    result = await client_multi.resource_list(tid=environment_multi, deploy_summary=True)
    assert result.code == 200
    assert result.result["metadata"]["deploy_summary"]["by_state"]["deployed"] == 1

    action_id = uuid.uuid4()
    now = datetime.now()
    result = await aclient.resource_action_update(
        environment_multi,
        ["std::testing::NullResource[vm1.dev.inmanta.com,name=hostname],v=%d" % version],
        action_id,
        "deploy",
        now,
        now,
        "deployed",
        [],
        {},
    )
    assert result.code == 200

    result = await client_multi.resource_list(tid=environment_multi, deploy_summary=True)
    assert result.code == 200
    assert result.result["metadata"]["deploy_summary"]["by_state"]["deployed"] == 2


async def test_get_environment(client, clienthelper, server, environment):
    for i in range(10):
        version = await clienthelper.get_version()

        resources = []
        for j in range(i):
            resources.append(
                {
                    "group": "root",
                    "hash": "89bf880a0dc5ffc1156c8d958b4960971370ee6a",
                    "id": "std::testing::NullResource[vm1.dev.inmanta.com,name=file%d],v=%d" % (j, version),
                    "owner": "root",
                    "path": "/tmp/file%d" % j,
                    "permissions": 644,
                    "purged": False,
                    "reload": False,
                    "requires": [],
                    "version": version,
                }
            )

        res = await client.put_version(
            tid=environment,
            version=version,
            resources=resources,
            unknowns=[],
            version_info={},
            compiler_version=get_compiler_version(),
        )
        assert res.code == 200

    result = await client.get_environment(environment, versions=5, resources=1)
    assert result.code == 200
    assert len(result.result["environment"]["versions"]) == 5
    assert len(result.result["environment"]["resources"]) == 9


async def test_resource_update(postgresql_client, client, clienthelper, server, environment, async_finalizer, null_agent):
    """
    Test updating resources and logging
    """

    aclient = null_agent._client

    version = await clienthelper.get_version()

    resources = []
    for j in range(10):
        resources.append(
            {
                "group": "root",
                "hash": "89bf880a0dc5ffc1156c8d958b4960971370ee6a",
                "id": "std::testing::NullResource[vm1,name=file%d],v=%d" % (j, version),
                "owner": "root",
                "path": "/tmp/file%d" % j,
                "permissions": 644,
                "purged": False,
                "reload": False,
                "requires": [],
                "version": version,
            }
        )

    res = await client.put_version(
        tid=environment,
        version=version,
        resources=resources,
        unknowns=[],
        version_info={},
        compiler_version=get_compiler_version(),
    )
    assert res.code == 200

    result = await client.release_version(environment, version, False)
    assert result.code == 200

    resource_ids = [x["id"] for x in resources]

    # Start the deploy
    action_id = uuid.uuid4()
    now = datetime.now()
    result = await aclient.resource_action_update(
        environment, resource_ids, action_id, "deploy", now, status=const.ResourceState.deploying
    )
    assert result.code == 200

    # Get the status from a resource
    result = await client.get_resource(tid=environment, id=resource_ids[0], logs=True)
    assert result.code == 200
    logs = {x["action"]: x for x in result.result["logs"]}

    assert "deploy" in logs
    assert logs["deploy"]["finished"] is None
    assert logs["deploy"]["messages"] is None
    assert logs["deploy"]["changes"] is None

    # Send some logs
    result = await aclient.resource_action_update(
        environment,
        resource_ids,
        action_id,
        "deploy",
        status=const.ResourceState.deploying,
        messages=[data.LogLine.log(const.LogLevel.INFO, "Test log %(a)s %(b)s", a="a", b="b")],
    )
    assert result.code == 200

    # Get the status from a resource
    result = await client.get_resource(tid=environment, id=resource_ids[0], logs=True)
    assert result.code == 200
    logs = {x["action"]: x for x in result.result["logs"]}

    assert "deploy" in logs
    assert "messages" in logs["deploy"]
    assert len(logs["deploy"]["messages"]) == 1
    assert logs["deploy"]["messages"][0]["msg"] == "Test log a b"
    assert logs["deploy"]["finished"] is None
    assert logs["deploy"]["changes"] is None

    # Finish the deploy
    now = datetime.now()
    changes = {x: {"owner": {"old": "root", "current": "inmanta"}} for x in resource_ids}
    result = await aclient.resource_action_update(environment, resource_ids, action_id, "deploy", finished=now, changes=changes)
    assert result.code == 400

    result = await aclient.resource_action_update(
        environment, resource_ids, action_id, "deploy", status=const.ResourceState.deployed, finished=now, changes=changes
    )
    assert result.code == 200
    assert await clienthelper.done_count() == 10


async def test_get_resource_on_invalid_resource_id(server, client, environment) -> None:
    """
    Verify that a clear error message is returned when the resource version id passed to the
    get_resource() endpoint has an invalid structure.
    """
    invalid_resource_version_id = "invalid resource version id"
    result = await client.get_resource(tid=environment, id=invalid_resource_version_id)
    assert result.code == 400
    assert f"{invalid_resource_version_id} is not a valid resource version id" in result.result["message"]


@pytest.mark.parametrize("no_agent", [True])
async def test_clear_environment(client, server, clienthelper, environment):
    """
    Test clearing out an environment
    """
    version = await clienthelper.get_version()
    result = await client.put_version(
        tid=environment, version=version, resources=[], unknowns=[], version_info={}, compiler_version=get_compiler_version()
    )
    assert result.code == 200

    result = await client.get_environment(id=environment, versions=10)
    assert result.code == 200
    assert len(result.result["environment"]["versions"]) == 1

    # trigger multiple compiles and wait for them to complete in order to test cascade deletion of collapsed compiles (#2350)
    result = await client.notify_change_get(id=environment)
    assert result.code == 200
    result = await client.notify_change_get(id=environment)
    assert result.code == 200
    result = await client.notify_change_get(id=environment)
    assert result.code == 200

    async def compile_done():
        return (await client.is_compiling(environment)).code == 204

    await retry_limited(compile_done, 10)

    # Wait for env directory to appear
    slice = server.get_slice(SLICE_SERVER)
    env_dir = os.path.join(slice._server_storage["server"], environment, "compiler")

    while not os.path.exists(env_dir):
        await asyncio.sleep(0.1)

    result = await client.clear_environment(id=environment)
    assert result.code == 200

    assert not os.path.exists(env_dir)

    result = await client.get_environment(id=environment, versions=10)
    assert result.code == 200
    assert len(result.result["environment"]["versions"]) == 0


async def test_tokens(server_multi, client_multi, environment_multi, request):
    # Test using API tokens

    # Check the parameters of the 'server_multi' fixture
    if request.node.callspec.id in ["SSL", "Normal"]:
        # Generating tokens is not allowed if auth is not enabled
        return

    test_token = client_multi._transport_instance.token
    token = await client_multi.create_token(environment_multi, ["api"], idempotent=True)
    jot = token.result["token"]

    assert jot != test_token

    client_multi._transport_instance.token = jot

    # try to access a non environment call (global)
    result = await client_multi.list_environments()
    assert result.code == 403

    result = await client_multi.list_versions(environment_multi)
    assert result.code == 200

    token = await client_multi.create_token(environment_multi, ["agent"], idempotent=True)
    agent_jot = token.result["token"]

    client_multi._transport_instance.token = agent_jot
    result = await client_multi.list_versions(environment_multi)
    assert result.code == 403


async def test_token_without_auth(server, client, environment):
    """Generating a token when auth is not enabled is not allowed"""
    token = await client.create_token(environment, ["api"], idempotent=True)
    assert token.code == 400


async def test_batched_code_upload(
    server_multi, client_multi, sync_client_multi, environment_multi, agent_multi, snippetcompiler
):
    """Test uploading all code definitions at once"""
    snippetcompiler.setup_for_snippet(
        """
    import std::testing
    f = std::testing::NullResource(name="test")
    """
    )
    version, _ = await snippetcompiler.do_export_and_deploy(do_raise=False)

    code_manager = loader.CodeManager()

    for type_name, resource_definition in resources.resource.get_resources():
        code_manager.register_code(type_name, resource_definition)

    for type_name, handler_definition in handler.Commander.get_providers():
        code_manager.register_code(type_name, handler_definition)

    await asyncio.get_event_loop().run_in_executor(
        None, lambda: upload_code(sync_client_multi, environment_multi, version, code_manager)
    )

    for name, source_info in code_manager.get_types():
        assert len(source_info) >= 2
        for info in source_info:
            # fetch the code from the server
            response = await agent_multi._client.get_file(info.hash)
            assert response.code == 200
            source_code = base64.b64decode(response.result["content"])
            assert info.content == source_code


@pytest.mark.parametrize("auto_start_agent", [True])
async def test_resource_action_log(server, client, environment, clienthelper, snippetcompiler):
    await clienthelper.set_auto_deploy()

    snippetcompiler.setup_for_snippet(
        """\
        import std::testing

        std::testing::NullResource(name="network", agentname="internal")
        """,
        autostd=True,
    )

    version, _ = await snippetcompiler.do_export_and_deploy()

    await clienthelper.wait_for_released(version)
    await clienthelper.wait_for_deployed(version)
    resource_action_log = os.path.join(config.log_dir.get(), f"{opt.server_resource_action_log_prefix.get()}{environment}.log")
    print(resource_action_log)
    logging.root.warning("%s", resource_action_log)

    log = os.path.join(config.log_dir.get(), f"agent-{environment}.log")
    with open(log, "r") as fh:
        for line in fh:
            logging.warning(line)

    assert os.path.isfile(resource_action_log)
    assert os.stat(resource_action_log).st_size != 0
    with open(resource_action_log) as f:
        contents = f.read()
        parts = contents.split(" ")
        # Date and time
        parser.parse(f"{parts[0]} {parts[1]}")


async def test_invalid_sid(server, client, environment):
    """
    Verify that API endpoints, that should only be called by an agent, return an HTTP 400
    if they are called without a session id.
    """
    res = await client.discovered_resource_create(
        tid=environment,
        discovered_resource_id="test::Test[agent1,attr=val]",
        discovery_resource_id="test::Test[agent1,attr=other_val]",
    )
    assert res.code == 400
    assert res.result["message"] == "Invalid request: this is an agent to server call, it should contain an agent session id"


@pytest.mark.parametrize("tz_aware_timestamp", [True, False])
async def test_get_param(server, client, environment, tz_aware_timestamp: bool):
    config.Config.set("server", "tz-aware-timestamps", str(tz_aware_timestamp).lower())

    metadata = {"key1": "val1", "key2": "val2"}

    await client.set_param(environment, "param", ParameterSource.user, "val", "", metadata, False)
    await client.set_param(environment, "param2", ParameterSource.user, "val2", "", {"a": "b"}, False)

    res = await client.list_params(tid=environment, query={"key1": "val1"})
    assert res.code == 200

    def check_datetime_serialization(timestamp: str, tz_aware_timestamp):
        """
        Check that the given timestamp was serialized appropriately according to the server.tz-aware-timestamps option.
        """
        expected_format: str = const.TIME_ISOFMT
        if tz_aware_timestamp:
            expected_format += "%z"
        is_aware = datetime.strptime(timestamp, expected_format).tzinfo is not None
        assert is_aware == tz_aware_timestamp

    check_datetime_serialization(res.result["now"], tz_aware_timestamp)

    parameters = res.result["parameters"]
    assert len(parameters) == 1

    metadata_received = parameters[0]["metadata"]
    assert len(metadata_received) == 2
    for k, v in metadata.items():
        assert k in metadata_received
        assert metadata_received[k] == v

    res = await client.list_params(tid=environment, query={})
    assert res.code == 200
    parameters = res.result["parameters"]
    assert len(parameters) == 2


async def test_server_logs_address(server_config, caplog, async_finalizer):
    ibl = InmantaBootloader(configure_logging=True)
    with caplog.at_level(logging.INFO):
        async_finalizer.add(partial(ibl.stop, timeout=15))
        await ibl.start()

        client = Client("client")
        result = await client.create_project("env-test")
        assert result.code == 200
        address = "127.0.0.1"

        log_contains(caplog, "protocol.rest", logging.INFO, f"Server listening on {address}:")


class MockConnection:
    """
    Mock connection class to simulate an asyncpg connection.
    This class includes a close method to mimic closing a database connection.
    """

    async def close(self, timeout: int) -> None:
        return


@pytest.mark.parametrize("db_wait_time", ["20", "0"])
async def test_bootloader_db_wait(monkeypatch, tmpdir, caplog, db_wait_time: str) -> None:
    """
    Tests the Inmanta server bootloader's behavior with respect to waiting for the database to be ready before proceeding
    with the startup, based on the 'db_wait_time' configuration.
    """
    state_dir: str = tmpdir.mkdir("state_dir").strpath
    config.Config.set("database", "wait_time", db_wait_time)
    config.Config.set("config", "state-dir", state_dir)

    state = {"first_connect": True}

    async def mock_asyncpg_connect(*args, **kwargs) -> MockConnection:
        """
        Mock function to replace asyncpg.connect.
        Will raise an Exception on the first invocation.
        """
        if state["first_connect"]:
            state["first_connect"] = False
            raise Exception("Connection failure")
        else:
            return MockConnection()

    async def mock_start(self) -> None:
        """Mocks the call to self.restserver.start()."""
        return

    monkeypatch.setattr("inmanta.server.protocol.Server.start", mock_start)
    monkeypatch.setattr("asyncpg.connect", mock_asyncpg_connect)
    ibl: InmantaBootloader = InmantaBootloader(configure_logging=True)
    caplog.set_level(logging.INFO)
    caplog.clear()
    start_task: asyncio.Task = asyncio.create_task(ibl.start())
    await start_task

    if db_wait_time != "0":
        log_contains(caplog, "inmanta.server.bootloader", logging.INFO, "Waiting for database to be up.")
        log_contains(caplog, "inmanta.server.bootloader", logging.INFO, "Successfully connected to the database.")
    else:
        # If db_wait_time is "0", the wait_for_db method is not called,
        # hence "Successfully connected to the database." log message will not appear.
        log_doesnt_contain(caplog, "inmanta.server.bootloader", logging.INFO, "Successfully connected to the database.")

    log_contains(caplog, "inmanta.server.server", logging.INFO, "Starting server endpoint")

    await ibl.stop(timeout=20)


@pytest.mark.parametrize("db_wait_time", ["2", "0"])
async def test_bootloader_connect_running_db(server_config, postgres_db, caplog, db_wait_time: str):
    """
    Tests that the bootloader can connect to a database and can start for both wait_up values
    """
    config.Config.set("database", "wait_time", db_wait_time)
    ibl: InmantaBootloader = InmantaBootloader(configure_logging=True)
    caplog.clear()
    caplog.set_level(logging.INFO)
    await ibl.start()
    await ibl.stop(timeout=20)

    if db_wait_time != "0":
        log_contains(caplog, "inmanta.server.bootloader", logging.INFO, "Successfully connected to the database.")
    else:
        # If db_wait_time is "0", the wait_for_db method is not called,
        # hence "Successfully connected to the database." log message will not appear.
        log_doesnt_contain(caplog, "inmanta.server.bootloader", logging.INFO, "Successfully connected to the database.")
    log_contains(caplog, "inmanta.server.server", logging.INFO, "Starting server endpoint")


async def test_get_resource_actions(postgresql_client, client, clienthelper, server, environment, null_agent):
    """
    Test querying resource actions via the API
    """
    aclient = null_agent._client

    agentmanager = server.get_slice(SLICE_AGENT_MANAGER)
    await retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    version = await clienthelper.get_version()

    resources = []
    for j in range(10):
        resources.append(
            {
                "group": "root",
                "hash": "89bf880a0dc5ffc1156c8d958b4960971370ee6a",
                "id": "std::testing::NullResource[vm1,name=file%d],v=%d" % (j, version),
                "owner": "root",
                "path": "/tmp/file%d" % j,
                "permissions": 644,
                "purged": False,
                "reload": False,
                "requires": [],
                "version": version,
            }
        )

    #  adding a resource action with its change field set to "created" to test the get_resource_actions
    #  filtering on resources with changes

    rvid_r1_v1 = f"std::testing::NullResource[agent1,name=file200],v={version}"
    resources.append(
        {
            "group": "root",
            "hash": "89bf880a0dc5ffc1156c8d958b4960971370ee6a",
            "id": rvid_r1_v1,
            "owner": "root",
            "path": "/tmp/file200",
            "permissions": 644,
            "purged": False,
            "reload": False,
            "requires": [],
            "version": version,
        }
    )

    res = await client.put_version(
        tid=environment,
        version=version,
        resources=resources,
        unknowns=[],
        version_info={},
        compiler_version=get_compiler_version(),
    )
    assert res.code == 200

    result = await client.release_version(environment, version, False)
    assert result.code == 200

    resource_ids_nochange = [x["id"] for x in resources[0:-1]]
    resource_ids_created = [resources[-1]["id"]]

    # Start the deploy
    action_id = uuid.uuid4()
    now = datetime.now().astimezone()
    result = await aclient.resource_action_update(
        environment,
        resource_ids_created,
        action_id,
        "deploy",
        now,
        status=const.ResourceState.deploying,
        change=const.Change.created,
    )
    assert result.code == 200

    action_id = uuid.uuid4()
    result = await aclient.resource_action_update(
        environment, resource_ids_nochange, action_id, "deploy", now, status=const.ResourceState.deploying
    )
    assert result.code == 200

    # Get the status from a resource
    result = await client.get_resource_actions(tid=environment)
    assert result.code == 200
    assert len(result.result["data"]) == 3

    result = await client.get_resource_actions(tid=environment, attribute="path")
    assert result.code == 400
    result = await client.get_resource_actions(tid=environment, attribute_value="/tmp/file")
    assert result.code == 400
    result = await client.get_resource_actions(tid=environment, attribute="path", attribute_value="/tmp/file1")
    assert result.code == 200
    assert len(result.result["data"]) == 2
    # Query actions happening earlier than the deploy
    result = await client.get_resource_actions(tid=environment, last_timestamp=now)
    assert result.code == 200
    assert len(result.result["data"]) == 1
    assert result.result["data"][0]["action"] == "store"
    # Query actions happening later than the start of the test case
    result = await client.get_resource_actions(tid=environment, first_timestamp=now - timedelta(minutes=1))
    assert result.code == 200
    assert len(result.result["data"]) == 3
    result = await client.get_resource_actions(tid=environment, first_timestamp=now - timedelta(minutes=1), last_timestamp=now)
    assert result.code == 400
    result = await client.get_resource_actions(tid=environment, action_id=action_id)
    assert result.code == 400
    result = await client.get_resource_actions(tid=environment, first_timestamp=now - timedelta(minutes=1), action_id=action_id)
    assert result.code == 200
    assert len(result.result["data"]) == 3

    exclude_changes = [const.Change.nochange.value, const.Change.created.value]
    result = await client.get_resource_actions(tid=environment, exclude_changes=exclude_changes)
    assert result.code == 200
    assert len(result.result["data"]) == 0

    exclude_changes = []
    result = await client.get_resource_actions(tid=environment, exclude_changes=exclude_changes)
    assert result.code == 200
    assert len(result.result["data"]) == 3

    exclude_changes = [const.Change.nochange.value]
    result = await client.get_resource_actions(tid=environment, exclude_changes=exclude_changes)
    assert result.code == 200
    assert len(result.result["data"]) == 1  # only one of the 3 resource_actions has change != nochange

    exclude_changes = ["error"]
    result = await client.get_resource_actions(tid=environment, exclude_changes=exclude_changes)
    assert result.code == 400
    assert "Failed to validate argument" in result.result["message"]


async def test_resource_action_pagination(postgresql_client, client, clienthelper, server, agent):
    """Test querying resource actions via the API, including the pagination links."""
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    # Add multiple versions of model
    for i in range(1, 12):
        cm = data.ConfigurationModel(
            environment=env.id,
            version=i,
            date=datetime.now(),
            total=1,
            version_info={},
            is_suitable_for_partial_compiles=False,
        )
        await cm.insert()
        res1 = data.Resource.new(
            environment=env.id,
            resource_version_id="std::testing::NullResource[agent1,name=motd],v=%s" % str(i),
            status=const.ResourceState.deployed,
            attributes={"attr": [{"a": 1, "b": "c"}], "path": "/etc/motd"},
        )
        await res1.insert()

    # Add resource actions for motd
    motd_first_start_time = datetime.now()
    earliest_action_id = uuid.uuid4()
    resource_action = data.ResourceAction(
        environment=env.id,
        version=1,
        resource_version_ids=[f"std::testing::NullResource[agent1,name=motd],v={1}"],
        action_id=earliest_action_id,
        action=const.ResourceAction.deploy,
        started=motd_first_start_time - timedelta(minutes=1),
    )
    await resource_action.insert()
    resource_action.add_logs([data.LogLine.log(logging.INFO, "Successfully stored version %(version)d", version=1)])
    await resource_action.save()

    action_ids_with_the_same_timestamp = []
    for i in range(2, 7):
        action_id = uuid.uuid4()
        action_ids_with_the_same_timestamp.append(action_id)
        resource_action = data.ResourceAction(
            environment=env.id,
            version=i,
            resource_version_ids=[f"std::testing::NullResource[agent1,name=motd],v={i}"],
            action_id=action_id,
            action=const.ResourceAction.deploy,
            started=motd_first_start_time,
        )
        await resource_action.insert()
        resource_action.add_logs([data.LogLine.log(logging.INFO, "Successfully stored version %(version)d", version=i)])
        await resource_action.save()
    action_ids_with_the_same_timestamp = sorted(action_ids_with_the_same_timestamp, reverse=True)
    later_action_id = uuid.uuid4()
    resource_action = data.ResourceAction(
        environment=env.id,
        version=6,
        resource_version_ids=[f"std::testing::NullResource[agent1,name=motd],v={6}"],
        action_id=later_action_id,
        action=const.ResourceAction.deploy,
        started=motd_first_start_time + timedelta(minutes=6),
    )
    await resource_action.insert()
    resource_action.add_logs([data.LogLine.log(logging.INFO, "Successfully stored version %(version)d", version=6)])
    await resource_action.save()

    result = await client.get_resource_actions(
        tid=env.id,
        resource_type="std::testing::NullResource",
        attribute="path",
        attribute_value="/etc/motd",
        last_timestamp=motd_first_start_time + timedelta(minutes=7),
        limit=2,
    )
    assert result.code == 200
    resource_actions = result.result["data"]
    expected_action_ids = [later_action_id] + action_ids_with_the_same_timestamp[:1]
    assert [uuid.UUID(resource_action["action_id"]) for resource_action in resource_actions] == expected_action_ids

    # Use the next link for pagination
    next_page = result.result["links"]["next"]
    port = opt.server_bind_port.get()
    base_url = f"http://localhost:{port}"
    url = f"{base_url}{next_page}"
    client = AsyncHTTPClient()
    request = HTTPRequest(
        url=url,
        headers={"X-Inmanta-tid": str(env.id)},
    )
    response = await client.fetch(request, raise_error=False)
    assert response.code == 200
    response = json.loads(response.body.decode("utf-8"))
    second_page_action_ids = [uuid.UUID(resource_action["action_id"]) for resource_action in response["data"]]
    assert second_page_action_ids == action_ids_with_the_same_timestamp[1:3]
    next_page = response["links"]["next"]
    url = f"{base_url}{next_page}"
    request.url = url
    response = await client.fetch(request, raise_error=False)
    assert response.code == 200
    response = json.loads(response.body.decode("utf-8"))
    third_page_action_ids = [uuid.UUID(resource_action["action_id"]) for resource_action in response["data"]]
    assert third_page_action_ids == action_ids_with_the_same_timestamp[3:5]
    # Go back to the previous page
    prev_page = response["links"]["prev"]
    url = f"{base_url}{prev_page}"
    request.url = url
    response = await client.fetch(request, raise_error=False)
    assert response.code == 200
    response = json.loads(response.body.decode("utf-8"))
    action_ids = [uuid.UUID(resource_action["action_id"]) for resource_action in response["data"]]
    assert action_ids == second_page_action_ids
    # And back to the third
    prev_page = response["links"]["next"]
    url = f"{base_url}{prev_page}"
    request.url = url
    response = await client.fetch(request, raise_error=False)
    assert response.code == 200
    response = json.loads(response.body.decode("utf-8"))
    action_ids = [uuid.UUID(resource_action["action_id"]) for resource_action in response["data"]]
    assert action_ids == third_page_action_ids


@pytest.mark.parametrize("no_agent", [True])
@pytest.mark.parametrize("method_to_use", ["send_in_progress", "resource_action_update"])
async def test_send_in_progress(server, client, environment, agent, method_to_use: str):
    """
    Ensure that the `ToDbUpdateManager.send_in_progress()` method and the `resource_action_update()` API endpoint do the same
    when a new deployment is reported.
    """
    env_id = uuid.UUID(environment)

    model_version = 1
    cm = data.ConfigurationModel(
        environment=env_id,
        version=model_version,
        date=datetime.now().astimezone(),
        total=1,
        version_info={},
        is_suitable_for_partial_compiles=False,
    )
    await cm.insert()

    model_version = 1
    rvid_r1 = "std::testing::NullResource[agent1,name=file1]"
    rvid_r2 = "std::testing::NullResource[agent1,name=file2]"
    rvid_r3 = "std::testing::NullResource[agent1,name=file3]"
    rvid_r1_v1 = f"{rvid_r1},v={model_version}"
    rvid_r2_v1 = f"{rvid_r2},v={model_version}"
    rvid_r3_v1 = f"{rvid_r3},v={model_version}"

    async def make_resource_with_last_non_deploying_status(
        status: const.ResourceState,
        last_non_deploying_status: const.NonDeployingResourceState,
        resource_version_id: str,
        attributes: dict[str, object],
        version: int,
    ) -> None:
        r1 = data.Resource.new(
            environment=env_id,
            status=status,
            resource_version_id=resource_version_id,
            attributes=attributes,
        )
        await r1.insert()
        await data.ResourcePersistentState.populate_for_version(environment=uuid.UUID(environment), model_version=version)
        await r1.update_persistent_state(last_deploy=datetime.now(tz=UTC), last_non_deploying_status=last_non_deploying_status)

    await make_resource_with_last_non_deploying_status(
        status=const.ResourceState.skipped,
        last_non_deploying_status=const.NonDeployingResourceState.skipped,
        resource_version_id=rvid_r1_v1,
        attributes={"purge_on_delete": False, "requires": [rvid_r2, rvid_r3]},
        version=model_version,
    )
    await make_resource_with_last_non_deploying_status(
        status=const.ResourceState.deployed,
        last_non_deploying_status=const.NonDeployingResourceState.deployed,
        resource_version_id=rvid_r2_v1,
        attributes={"purge_on_delete": False, "requires": []},
        version=model_version,
    )
    await make_resource_with_last_non_deploying_status(
        status=const.ResourceState.failed,
        last_non_deploying_status=const.NonDeployingResourceState.failed,
        resource_version_id=rvid_r3_v1,
        attributes={"purge_on_delete": False, "requires": []},
        version=model_version,
    )

    action_id = uuid.uuid4()

    if method_to_use == "send_in_progress":
        update_manager = persistence.ToDbUpdateManager(client, env_id)
        await update_manager.send_in_progress(action_id, resources.Id.parse_id(rvid_r1_v1))
    else:
        await agent._client.resource_action_update(
            tid=env_id,
            resource_ids=[rvid_r1_v1],
            action_id=action_id,
            action=const.ResourceAction.deploy,
            started=datetime.now().astimezone(),
            status=const.ResourceState.deploying,
        )

    # Ensure that both API calls result in the same behavior
    result = await client.get_resource_actions(tid=env_id)
    assert result.code == 200
    assert len(result.result["data"]) == 1
    resource_action = result.result["data"][0]
    assert resource_action["environment"] == str(env_id)
    assert resource_action["version"] == model_version
    assert resource_action["resource_version_ids"] == [rvid_r1_v1]
    assert resource_action["action_id"] == str(action_id)
    assert resource_action["action"] == const.ResourceAction.deploy
    assert resource_action["started"] is not None
    assert resource_action["finished"] is None
    assert resource_action["status"] == const.ResourceState.deploying
    assert resource_action["changes"] is None
    assert resource_action["change"] is None


async def test_send_in_progress_action_id_conflict(server, client, environment, agent):
    """
    Ensure proper error handling when the same action_id is provided twice to the `ToDbUpdateManager.send_in_progress()` method.
    """
    env_id = uuid.UUID(environment)

    model_version = 1
    cm = data.ConfigurationModel(
        environment=env_id,
        version=model_version,
        date=datetime.now().astimezone(),
        total=1,
        version_info={},
        is_suitable_for_partial_compiles=False,
    )
    await cm.insert()

    model_version = 1
    rvid_r1_v1 = ResourceVersionIdStr(f"std::testing::NullResource[agent1,name=file1],v={model_version}")

    await data.Resource.new(
        environment=env_id,
        status=const.ResourceState.skipped,
        resource_version_id=rvid_r1_v1,
        attributes={"purge_on_delete": False, "requires": []},
    ).insert()

    action_id = uuid.uuid4()
    update_manager = persistence.ToDbUpdateManager(client, env_id)

    async def execute_send_in_progress(expect_exception: bool, resulting_nr_resource_actions: int) -> None:
        try:
            await update_manager.send_in_progress(action_id, resources.Id.parse_id(rvid_r1_v1))
        except ValueError:
            assert expect_exception
        else:
            assert not expect_exception

        result = await client.get_resource_actions(tid=env_id)
        assert result.code == 200
        assert len(result.result["data"]) == resulting_nr_resource_actions

    await execute_send_in_progress(expect_exception=False, resulting_nr_resource_actions=1)
    await execute_send_in_progress(expect_exception=True, resulting_nr_resource_actions=1)


@pytest.mark.parametrize(
    "method_to_use",
    ["send_deploy_done", "resource_action_update"],
)
async def test_send_deploy_done(server, client, environment, null_agent, caplog, method_to_use, clienthelper):
    """
    Ensure that the `send_deploy_done` method behaves in the same way as the `resource_action_update` endpoint
    when the finished field is not None.
    """
    result = await client.set_setting(environment, "auto_deploy", True)
    assert result.code == 200

    env_id = uuid.UUID(environment)
    model_version = await clienthelper.get_version()
    rid_r1 = ResourceIdStr("std::testing::NullResource[agent1,name=file1]")
    rvid_r1_v1 = ResourceVersionIdStr(f"{rid_r1},v={model_version}")
    attributes_r1 = {
        "name": "file1",
        "id": f"std::testing::NullResource[agent1,name=file1],v={model_version}",
        "version": model_version,
        "purge_on_delete": False,
        "purged": True,
        "requires": [],
    }
    await clienthelper.put_version_simple(resources=[attributes_r1], version=model_version, wait_for_released=True)

    # Add parameter for resource
    parameter_id = "test_param"
    result = await client.set_param(
        tid=env_id,
        id=parameter_id,
        source=const.ParameterSource.user,
        value="val",
        resource_id="std::testing::NullResource[agent1,name=file1]",
    )
    assert result.code == 200

    update_manager = persistence.ToDbUpdateManager(client, env_id)
    action_id = uuid.uuid4()
    await update_manager.send_in_progress(action_id, resources.Id.parse_id(rvid_r1_v1))

    # Assert initial state
    result = await client.get_resource_actions(tid=env_id)
    assert result.code == 200, result.result
    deploy_resource_actions = [r for r in result.result["data"] if r["action"] == const.ResourceAction.deploy.value]
    assert len(deploy_resource_actions) == 1
    resource_action = deploy_resource_actions[0]
    assert resource_action["environment"] == str(env_id)
    assert resource_action["version"] == model_version
    assert resource_action["resource_version_ids"] == [rvid_r1_v1]
    assert resource_action["action_id"] == str(action_id)
    assert resource_action["action"] == const.ResourceAction.deploy
    assert resource_action["started"] is not None
    assert resource_action["finished"] is None
    assert resource_action["status"] == const.ResourceState.deploying
    assert resource_action["changes"] is None
    assert resource_action["change"] is None

    result = await client.get_resource(tid=env_id, id=rvid_r1_v1)
    assert result.code == 200, result.result

    caplog.clear()
    with caplog.at_level(logging.DEBUG):
        # Mark deployment as done
        now = datetime.now().astimezone()
        messages = [
            data.LogLine.log(level=const.LogLevel.DEBUG, msg="message", timestamp=now, keyword=123, none=None),
            data.LogLine.log(level=const.LogLevel.INFO, msg="test", timestamp=now),
        ]
        if method_to_use == "send_deploy_done":
            await update_manager.send_deploy_done(
                attribute_hash=util.make_attribute_hash(resource_id=rid_r1, attributes=attributes_r1),
                result=executor.DeployReport(
                    rvid=rvid_r1_v1,
                    action_id=action_id,
                    resource_state=const.HandlerResourceState.deployed,
                    messages=messages,
                    changes={"attr1": AttributeStateChange(current=None, desired="test")},
                    change=const.Change.purged,
                ),
                state=state.ResourceState(
                    compliance=state.Compliance.COMPLIANT,
                    last_deploy_result=state.DeployResult.DEPLOYED,
                    blocked=state.Blocked.NOT_BLOCKED,
                    last_deployed=now,
                ),
                started=now,
                finished=now,
            )
        else:
            result = await null_agent._client.resource_action_update(
                tid=env_id,
                resource_ids=[rvid_r1_v1],
                action_id=action_id,
                action=const.ResourceAction.deploy,
                started=None,
                finished=now,
                status=const.ResourceState.deployed,
                messages=messages,
                changes={rvid_r1_v1: {"attr1": AttributeStateChange(current=None, desired="test")}},
                change=const.Change.purged,
                send_events=True,
            )
            assert result.code == 200, result.result

    result = await client.get_resource_actions(tid=env_id)
    assert result.code == 200, result.result
    deploy_resource_actions = [r for r in result.result["data"] if r["action"] == const.ResourceAction.deploy.value]
    assert len(deploy_resource_actions) == 1
    resource_action = deploy_resource_actions[0]
    assert resource_action["environment"] == str(env_id)
    assert resource_action["version"] == model_version
    assert resource_action["resource_version_ids"] == [rvid_r1_v1]
    assert resource_action["action_id"] == str(action_id)
    assert resource_action["action"] == const.ResourceAction.deploy
    assert resource_action["started"] is not None
    assert resource_action["finished"] is not None

    expected_timestamp: str
    if opt.server_tz_aware_timestamps.get():
        expected_timestamp = now.astimezone().isoformat(timespec="microseconds")
    else:
        expected_timestamp = now.astimezone(timezone.utc).replace(tzinfo=None).isoformat()

    expected_resource_action_messages = [
        {
            "level": const.LogLevel.DEBUG.name,
            "msg": "message",
            "args": [],
            "kwargs": {"keyword": 123, "none": None},
            "timestamp": expected_timestamp,
        },
        {
            "level": const.LogLevel.INFO.name,
            "msg": "test",
            "args": [],
            "kwargs": {},
            "timestamp": expected_timestamp,
        },
    ]
    assert resource_action["messages"] == expected_resource_action_messages
    assert resource_action["status"] == const.ResourceState.deployed
    assert resource_action["changes"] == {rvid_r1_v1: {"attr1": AttributeStateChange(current=None, desired="test").dict()}}
    assert resource_action["change"] == const.Change.purged.value

    result = await client.resource_details(tid=env_id, rid=rid_r1)
    assert result.code == 200, result.result
    assert result.result["data"]["status"] == const.ResourceState.deployed

    result = await client.get_version(tid=env_id, id=1)
    assert result.code == 200, result.result

    # parameter was deleted due to purge operation
    result = await client.list_params(tid=env_id)
    assert result.code == 200
    assert len(result.result["parameters"]) == 0

    # A new send_deploy_done call for the same action_id should result in a ValueError
    with pytest.raises(ValueError):
        await update_manager.send_deploy_done(
            attribute_hash=util.make_attribute_hash(resource_id=rid_r1, attributes=attributes_r1),
            result=executor.DeployReport(
                rvid=rvid_r1_v1,
                action_id=action_id,
                resource_state=const.HandlerResourceState.deployed,
                messages=[],
                changes={"attr1": AttributeStateChange(current="test", desired="test2")},
                change=const.Change.created,
            ),
            state=state.ResourceState(
                compliance=state.Compliance.COMPLIANT,
                last_deploy_result=state.DeployResult.DEPLOYED,
                blocked=state.Blocked.NOT_BLOCKED,
                last_deployed=datetime.now().astimezone(),
            ),
            started=datetime.now().astimezone(),
            finished=datetime.now().astimezone(),
        )


async def test_send_deploy_done_error_handling(server, client, environment, agent):
    env_id = uuid.UUID(environment)
    update_manager = persistence.ToDbUpdateManager(client, uuid.UUID(environment))

    model_version = 1
    cm = data.ConfigurationModel(
        environment=env_id,
        version=model_version,
        date=datetime.now().astimezone(),
        total=1,
        version_info={},
        is_suitable_for_partial_compiles=False,
    )
    await cm.insert()

    rvid_r1_v1 = ResourceVersionIdStr(f"std::testing::NullResource[agent1,name=file1],v={model_version}")

    # Resource doesn't exist
    with pytest.raises(ValueError) as exec_info:
        await update_manager.send_deploy_done(
            attribute_hash="",
            result=executor.DeployReport(
                rvid=rvid_r1_v1,
                action_id=uuid.uuid4(),
                resource_state=const.HandlerResourceState.deployed,
                messages=[],
                changes={},
                change=const.Change.nochange,
            ),
            state=state.ResourceState(
                compliance=state.Compliance.COMPLIANT,
                last_deploy_result=state.DeployResult.DEPLOYED,
                blocked=state.Blocked.NOT_BLOCKED,
                last_deployed=datetime.now().astimezone(),
            ),
            started=datetime.now().astimezone(),
            finished=datetime.now().astimezone(),
        )
    assert "The resource with the given id does not exist in the given environment" in str(exec_info.value)

    # Create resource
    await data.Resource.new(
        environment=env_id,
        status=const.ResourceState.available,
        resource_version_id=rvid_r1_v1,
        attributes={"purge_on_delete": False, "requires": []},
    ).insert()

    # Resource action doesn't exist
    with pytest.raises(ValueError) as exec_info:
        await update_manager.send_deploy_done(
            attribute_hash="",
            result=executor.DeployReport(
                rvid=rvid_r1_v1,
                action_id=uuid.uuid4(),
                resource_state=const.HandlerResourceState.deployed,
                messages=[],
                changes={},
                change=const.Change.nochange,
            ),
            state=state.ResourceState(
                compliance=state.Compliance.COMPLIANT,
                last_deploy_result=state.DeployResult.DEPLOYED,
                blocked=state.Blocked.NOT_BLOCKED,
                last_deployed=datetime.now().astimezone(),
            ),
            started=datetime.now().astimezone(),
            finished=datetime.now().astimezone(),
        )
    assert "No resource action exists for action_id" in str(exec_info.value)


async def test_start_location_no_redirect(server):
    """
    Ensure that there is no redirection for the "start" location. (issue #3497)
    """
    port = opt.server_bind_port.get()
    base_url = f"http://localhost:{port}/"
    http_client = AsyncHTTPClient()
    request = HTTPRequest(
        url=base_url,
    )
    response = await http_client.fetch(request, raise_error=False)
    assert base_url == response.effective_url


@pytest.mark.parametrize("path", ["", "/", "/test"])
async def test_redirect_dashboard_to_console(server, path):
    """
    Ensure that there is a redirection from the dashboard to the webconsole
    """
    port = opt.server_bind_port.get()
    base_url = f"http://localhost:{port}/dashboard{path}"
    result_url = f"http://localhost:{port}/console{path}"
    http_client = AsyncHTTPClient()
    request = HTTPRequest(
        url=base_url,
    )
    response = await http_client.fetch(request, raise_error=False)
    assert result_url == response.effective_url


@pytest.mark.parametrize("env1_halted", [True, False])
@pytest.mark.parametrize("env2_halted", [True, False])
async def test_cleanup_old_agents(server, client, env1_halted, env2_halted):
    """
    This test is testing the functionality of cleaning up old agents in the database.
    The test creates 2 environments and adds agents with various properties (some used in a version
    and some with the primary ID set), and then tests that the cleanup function correctly removes
    only the agents that meet the criteria for deletion. Also verifies that only agents in envs that
    are not halted are cleaned up
    """

    project = data.Project(name="test")
    await project.insert()

    env1 = data.Environment(name="env1", project=project.id)
    await env1.insert()
    env2 = data.Environment(name="env2", project=project.id)
    await env2.insert()

    if env1_halted:
        result = await client.halt_environment(env1.id)
        assert result.code == 200
    if env2_halted:
        result = await client.halt_environment(env2.id)
        assert result.code == 200

    process_sid = uuid.uuid4()
    await data.AgentProcess(hostname="localhost-dummy", environment=env1.id, sid=process_sid, last_seen=datetime.now()).insert()

    id_primary = uuid.uuid4()
    await data.AgentInstance(id=id_primary, process=process_sid, name="dummy-instance", tid=env1.id).insert()

    version = 1
    await data.ConfigurationModel(
        environment=env1.id,
        version=version,
        date=datetime.now(),
        total=1,
        released=True,
        version_info={},
        is_suitable_for_partial_compiles=False,
    ).insert()

    name = "file1"
    resource_id = f"std::testing::NullResource[agent4,name={name}]"

    await data.Resource.new(
        environment=env1.id, resource_version_id=ResourceVersionIdStr(f"{resource_id},v={version}"), attributes={"name": name}
    ).insert()

    # should get purged
    await data.Agent(
        environment=env1.id,
        name="agent1",
        paused=False,
        id_primary=None,
    ).insert()
    # should not get purged as the id_primary is set -> not down
    await data.Agent(environment=env1.id, name="agent2", paused=False, id_primary=id_primary).insert()
    # should not get purged as it is used in a version of the ConfigurationModel
    await data.Agent(
        environment=env1.id,
        name="agent4",
        paused=False,
        id_primary=None,
    ).insert()
    # agent with "agent2" as name but in another env will get purged:
    await data.Agent(
        environment=env2.id,
        name="agent2",
        paused=False,
        id_primary=None,
    ).insert()
    # agent with "agent1" as name but in another env will get purged:
    await data.Agent(
        environment=env2.id,
        name="agent1",
        paused=False,
        id_primary=None,
    ).insert()

    agents_before_purge = await data.Agent.get_list()
    assert len(agents_before_purge) == 5

    await server.get_slice(SLICE_ORCHESTRATION)._purge_versions()

    agents_after_purge = [(agent.environment, agent.name) for agent in await data.Agent.get_list()]
    number_agents_env1_after_purge = 3 if env1_halted else 2
    number_agents_env2_after_purge = 2 if env2_halted else 0
    assert len(agents_after_purge) == number_agents_env1_after_purge + number_agents_env2_after_purge
    if not (env1_halted or env2_halted):
        expected_agents_after_purge = [
            (env1.id, "agent2"),
            (env1.id, "agent4"),
        ]
        assert sorted(agents_after_purge) == sorted(expected_agents_after_purge)


async def test_serialization_attributes_of_resource_to_api(client, server, environment, clienthelper, null_agent) -> None:
    """
    Due to a bug, the version of a resource was always included in the attribute dictionary.
    This issue has been patched in the database, but at the API boundary we still serve the version
    field in the attributes dictionary for backwards compatibility. This test verifies that behavior.
    """
    version = await clienthelper.get_version()
    resource_id = "test::Resource[agent1,key=key1]"
    resources = [
        {
            "id": f"{resource_id},v={version}",
            "att": "val",
            "version": version,
            "send_event": False,
            "purged": False,
            "requires": [],
        }
    ]
    attributes_on_api = {k: v for k, v in resources[0].items() if k != "id"}
    result = await client.put_version(
        tid=environment,
        version=version,
        resources=resources,
        unknowns=[],
        version_info={},
        compiler_version=get_compiler_version(),
    )
    assert result.code == 200

    result = await client.release_version(tid=environment, id=version)
    assert result.code == 200

    # Verify that the version field is not present in the attributes dictionary in the database.
    result = await data.Resource.get_list()
    assert len(result) == 1
    resource_dao = result[0]
    assert "version" not in resource_dao.attributes

    # Ensure that the serialization of the resource DAO contains the version field in the attributes dictionary
    resource_dto = resource_dao.to_dto()
    assert resource_dto.attributes["version"] == version
    resource_dct = resource_dao.to_dict()
    assert resource_dct["attributes"]["version"] == version

    # Retrieve the resource via the API and ensure that the version field is present in the attributes dictionary
    result = await client.resource_history(environment, resource_id)
    assert result.code == 200
    assert len(result.result["data"]) == 1
    assert result.result["data"][0]["attributes"] == attributes_on_api

    result = await client.versioned_resource_details(tid=environment, version=version, rid=resource_id)
    assert result.code == 200
    assert result.result["data"]["attributes"] == attributes_on_api, result.result["data"]

    result = await client.resource_details(tid=environment, rid=resource_id)
    assert result.code == 200
    assert result.result["data"]["attributes"] == attributes_on_api


@pytest.mark.parametrize("v1_partial,v2_partial", [(False, False), (False, True)])
# the other two cases require a race condition to trigger
# as put_partial determines its own version number
async def test_put_stale_version(client, server, environment, clienthelper, caplog, v1_partial, v2_partial):
    """Put a version in with auto deploy on that is already stale"""
    await client.set_setting(environment, AUTO_DEPLOY, True)

    v0 = await clienthelper.get_version()
    v1 = await clienthelper.get_version()
    v2 = await clienthelper.get_version()

    async def put_version(version: int) -> int:
        partial = (version == v1 and v1_partial) or (version == v2 and v2_partial)

        if partial:
            version = 0

        resource_id = "test::Resource[agent1,key=key1]"
        resources = [
            {
                "id": f"{resource_id},v={version}",
                "att": "val",
                "version": version,
                "send_event": False,
                "purged": False,
                "requires": [],
            }
        ]

        if partial:
            result = await client.put_partial(
                tid=environment,
                resources=resources,
                unknowns=[],
                version_info={},
            )
            assert result.code == 200
            return result.result["data"]
        else:
            result = await client.put_version(
                tid=environment,
                version=version,
                resources=resources,
                unknowns=[],
                version_info={},
                compiler_version=get_compiler_version(),
            )
            assert result.code == 200
            return version

    v0 = await put_version(v0)
    await retry_limited(functools.partial(clienthelper.is_released, v0), timeout=1, interval=0.05)
    v2 = await put_version(v2)
    await retry_limited(functools.partial(clienthelper.is_released, v2), timeout=1, interval=0.05)
    v1 = await put_version(v1)
    # give it time to attempt to be release
    await asyncio.sleep(0.1)
    assert not await clienthelper.is_released(v1)


async def test_set_fact_v2(
    server,
    client,
    clienthelper,
    environment,
):
    """
    Test the set_fact endpoint. First create a fact with expires set to true.
    Then set expires to false for the same fact.
    """
    version = await clienthelper.get_version()
    resource_id = "test::MyDiscoveryResource[discovery_agent,key=key1]"
    resource_version_id = f"{resource_id},v={version}"

    resources = [
        {
            "key": "key1",
            "id": resource_version_id,
            "send_event": True,
            "purged": False,
            "requires": [],
        }
    ]

    # Put a new version containing a resource with id=resource_id, to make sure the fact is not cleaned up.
    result = await client.put_version(
        tid=environment,
        version=version,
        resources=resources,
        unknowns=[],
        version_info={},
        compiler_version=util.get_compiler_version(),
    )
    assert result.code == 200

    result = await client.set_fact(
        tid=environment,
        name="test",
        source=ParameterSource.fact.value,
        value="value1",
        resource_id="test::MyDiscoveryResource[discovery_agent,key=key1]",
    )

    assert result.code == 200
    fact = result.result["data"]
    assert fact["expires"] is True

    result = await client.get_facts(
        tid=environment,
        rid="test::MyDiscoveryResource[discovery_agent,key=key1]",
    )
    assert result.code == 200
    assert len(result.result["data"]) == 1
    assert result.result["data"][0] == fact

    result = await client.set_fact(
        tid=environment,
        name="test",
        source=ParameterSource.fact.value,
        value="value1",
        resource_id="test::MyDiscoveryResource[discovery_agent,key=key1]",
        expires=False,
    )
    assert result.code == 200
    fact = result.result["data"]
    assert fact["expires"] is False

    result = await client.get_facts(
        tid=environment,
        rid="test::MyDiscoveryResource[discovery_agent,key=key1]",
    )
    assert result.code == 200
    assert len(result.result["data"]) == 1
    assert result.result["data"][0] == fact


async def test_set_param_v2(server, client, environment):
    """
    Test the set_parameter endpoint. Create a parameters and verify that expires is set to false.
    Also test we can modify it and create a second one.
    """

    result = await client.set_parameter(
        tid=environment,
        name="param",
        source=ParameterSource.user,
        value="val",
        metadata={"key1": "val1", "key2": "val2"},
        recompile=False,
    )

    assert result.code == 200

    res = await client.list_params(tid=environment, query={})
    assert res.code == 200
    parameters = res.result["parameters"]
    assert len(parameters) == 1
    assert parameters[0]["name"] == "param"
    assert parameters[0]["value"] == "val"
    assert parameters[0]["expires"] is False

    await client.set_parameter(
        tid=environment,
        name="param",
        source=ParameterSource.user,
        value="val2",
        metadata={"key1": "val1", "key2": "val2"},
        recompile=False,
    )
    assert result.code == 200

    res = await client.list_params(tid=environment, query={})
    assert res.code == 200
    parameters = res.result["parameters"]
    assert len(parameters) == 1
    assert parameters[0]["name"] == "param"
    assert parameters[0]["value"] == "val2"
    assert parameters[0]["expires"] is False

    await client.set_parameter(
        tid=environment, name="param2", source=ParameterSource.user, value="val3", metadata={}, recompile=False
    )
    assert result.code == 200

    res = await client.list_params(tid=environment, query={})
    assert res.code == 200
    parameters = res.result["parameters"]
    assert len(parameters) == 2


async def test_delete_active_version(client, clienthelper, server, environment, null_agent):
    """
    Test that the active version cannot be deleted
    """
    version = await clienthelper.get_version()
    assert version == 1
    res1 = "test::Resource[agent1,key=key1]"
    res2 = "test::Resource[agent1,key=key2]"
    resources = [
        {"key": "key1", "value": "value", "id": f"{res1},v={version}", "requires": [], "purged": False, "send_event": False},
        {"key": "key2", "value": "value", "id": f"{res2},v={version}", "requires": [], "purged": False, "send_event": False},
    ]

    await clienthelper.put_version_simple(resources, version)

    result = await client.release_version(
        environment, version, push=False, agent_trigger_method=const.AgentTriggerMethod.push_full_deploy
    )
    assert result.code == 200

    # Remove version 1
    result = await client.delete_version(tid=environment, id=version)
    assert result.code == 400
    assert result.result["message"] == "Invalid request: Cannot delete the active version"



"""
Copyright 2014 Omer Gertel
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com

This code was originally developed by Omer Gertel, as a python port of the core portion of a
[Java Metrics library by Coda Hale](http://metrics.dropwizard.io/)

It was vendored into the inmanta source tree as the original was no longer maintained.
"""

from threading import Lock


class Counter(object):
    """
    An incrementing and decrementing metric
    """

    def __init__(self) -> None:
        super(Counter, self).__init__()
        self.lock = Lock()
        self.counter = 0

    def inc(self, val: int = 1) -> None:
        "increment counter by val (default is 1)"
        with self.lock:
            self.counter = self.counter + val

    def dec(self, val: int = 1) -> None:
        "decrement counter by val (default is 1)"
        self.inc(-val)

    def get_count(self) -> int:
        "return current value of counter"
        return self.counter

    def clear(self) -> None:
        "reset counter to 0"
        with self.lock:
            self.counter = 0



"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import heapq
import logging
import sys
import time
from threading import Lock
from types import TracebackType
from typing import TYPE_CHECKING, Any, Callable, Optional, Type

from inmanta.resources import Resource
from inmanta.stable_api import stable_api

if TYPE_CHECKING:
    from inmanta.agent.executor import AgentInstance

LOGGER = logging.getLogger(__name__)


class CacheItem:
    def __init__(
        self,
        key: str,
        value: Any,
        call_on_delete: Optional[Callable[[Any], None]],
        evict_after_last_access: float,
        evict_after_creation: float,
    ) -> None:
        """
        :param key: The full key identifying this item in the cache.
        :param value: The value being cached associated to the key.
        :param call_on_delete: Optional finalizer to call when the cache item is deleted. This is
            a callable expecting the cached value as an argument.
        :param evict_after_last_access: This cache item will be considered stale this number of seconds after
            it was last accessed.
        :param evict_after_creation: This cache item will be considered stale this number of seconds after
            entering the cache.
        """
        self.key = key
        self.value = value
        self.call_on_delete = call_on_delete
        self.refresh_after_access = evict_after_last_access

        now = time.time()
        self.expiry_time: float = sys.float_info.max
        self.after_creation_expiry_time: float = sys.float_info.max

        if evict_after_last_access > 0:
            self.expiry_time = now + evict_after_last_access
        if evict_after_creation > 0:
            self.after_creation_expiry_time = now + evict_after_creation
            self.expiry_time = min(self.expiry_time, self.after_creation_expiry_time)

        # Make sure finalizers are only called once
        self.finalizer_lock = Lock()
        self.called_finalizer = False

    def __lt__(self, other: "CacheItem") -> bool:
        return self.expiry_time < other.expiry_time

    def delete(self) -> None:
        with self.finalizer_lock:
            if callable(self.call_on_delete) and not self.called_finalizer:
                self.called_finalizer = True
                self.call_on_delete(self.value)

    def __del__(self) -> None:
        self.delete()

    def refresh(self, now: float) -> None:
        """
        Refresh this cache item. Resets the expiry time for items marked
        for eviction a certain time after their last access.

        :parameter now: Baseline 'now' time to make sure expiry times are synced up
            across cache items.
        """
        if self.refresh_after_access <= 0:
            # Refreshing on access is disabled on the CacheItem.
            return
        self.expiry_time = min(now + self.refresh_after_access, self.after_creation_expiry_time)


@stable_api
class AgentCache:
    """
    Caching system for the agent.

    Cached items will live in the cache as long as they are not stale.
    This expiry time can be set with relation to item creation timestamp and/or
    item last access timestamp. If both are set, the shortest duration of the
    two will trigger the item to become stale.

    1. evict_after_last_access

        These items are expected to be reused across multiple
        model versions. Their expiry time is reset to the value
        set by the `evict_after_last_access` parameter of the
        `@cache` decorator any time they're accessed.

    2. evict_after_creation

        These items have a fixed lifetime. Their expiry time
        is set by the `evict_after_creation` parameter of the `@cache` decorator.

    To enforce consistency, this cache should be used as a context manager
    by the agent when performing a resource action (deploy / repair / fact retrieval / dry run).

    This ensures that:
        - before using the cache: we clean up stale entries.
        - when done using the cache: we reset the expiry time of items meant
          to be reused across multiple model versions.
    """

    def __init__(self, agent_instance: Optional["AgentInstance"] = None) -> None:
        """
        :param agent_instance: The AgentInstance that is using the cache. The value is None when the cache
                               is used from pytest-inmanta.
        """
        # The cache itself
        self.cache: dict[str, CacheItem] = {}

        # Time-based eviction mechanism
        # Keep track of when is the next earliest cache item expiry time.
        self.next_action: Optional[float] = None
        # Heap queue of cache items, used for efficient retrieval of the cache
        # item that expires the soonest. should only be mutated via the heapq API.
        self.timer_queue: list[CacheItem] = []

        self.addLock = Lock()
        self.addLocks: dict[str, Lock] = {}
        self._agent_instance = agent_instance

        # This set holds cache items used during a resource action whose
        # expiry time should be refreshed i.e. evict_after_last_access>0
        self.used_items_to_refresh: set[CacheItem] = set()

    def touch_used_cache_items(self) -> None:
        """
        Extend the expiry time of items in the used_items_to_refresh by their
        respective grace period.
        """
        now = time.time()
        for item in self.used_items_to_refresh:
            item.refresh(now)

        self.used_items_to_refresh = set()
        if self.timer_queue:
            # Bad O(N) but unavoidable (?) Since we modify elements above
            # we have to make sure the heap is still a heap.
            heapq.heapify(self.timer_queue)
            self.next_action = self.timer_queue[0].expiry_time

    def close(self) -> None:
        """
        Cleanly terminate the cache
        """
        self.next_action = None
        for key in list(self.cache.keys()):
            self._evict_item(key)
        self.timer_queue.clear()

    def _evict_item(self, key: str) -> None:
        """
        Evict an item from the cache by key.
        """
        try:
            item = self.cache[key]
            item.delete()
            del self.cache[key]
        except KeyError:
            # already gone
            pass

    def clean_stale_entries(self) -> None:
        """
        Remove stale entries from the cache.
        """
        now = time.time()
        while self.next_action is not None and now > self.next_action and len(self.timer_queue) > 0:
            item = heapq.heappop(self.timer_queue)
            self._evict_item(item.key)
            if len(self.timer_queue) > 0:
                self.next_action = self.timer_queue[0].expiry_time
            else:
                self.next_action = None

    def _get(self, key: str) -> CacheItem:
        """
        Retrieve cache item with the given key

        :param key: Key of the item being retrieved from the cache
        :return: The cached item

        :raises KeyError: If the key is not present in the cache
        """
        item = self.cache[key]

        if item.refresh_after_access:
            self.used_items_to_refresh.add(item)

        return item

    def _cache(self, item: CacheItem) -> None:
        if item.key in self.cache:
            raise Exception("Added same item twice")

        self.cache[item.key] = item

        heapq.heappush(self.timer_queue, item)

        if item.refresh_after_access:
            self.used_items_to_refresh.add(item)

        if self.next_action is None or item.expiry_time < self.next_action:
            self.next_action = item.expiry_time

    def _get_key(self, key: str, resource: Optional[Resource]) -> str:
        key_parts = [key]
        if resource is not None:
            key_parts.append(str(resource.id.resource_str()))
        return "__".join(key_parts)

    def cache_value(
        self,
        key: str,
        value: Any,
        evict_after_last_access: float = 60,
        evict_after_creation: float = 0,
        resource: Optional[Resource] = None,
        call_on_delete: Optional[Callable[[Any], None]] = None,
    ) -> None:
        """
        add a value to the cache with the given key

        if a resource is given, it is appended to the key

        :param key: Key for this item
        :param value: The value to cache
        :param evict_after_last_access: This cache item will be considered stale this number of seconds after
            it was last accessed.
        :param evict_after_creation: This cache item will be considered stale this number of seconds after
            entering the cache.
        :param resource: The resource associated with this entry
        :param call_on_delete: A callback function that is called when the value is removed from the cache.
        """
        self._cache(
            CacheItem(
                self._get_key(key, resource),
                value,
                call_on_delete,
                evict_after_last_access=evict_after_last_access,
                evict_after_creation=evict_after_creation,
            )
        )

    def find(self, key: str, resource: Optional[Resource] = None) -> Any:
        """
        find a value in the cache with the given key

        if a resource is given, it is appended to the key

        :raise KeyError: if the value is not found
        """
        item = self._get(self._get_key(key, resource)).value

        return item

    def get_or_else(
        self,
        key: str,
        function: Callable[..., Any],
        # deprecated parameter, kept for backwards compatibility
        for_version: Optional[bool] = None,
        # deprecated parameter, kept for backwards compatibility
        timeout: Optional[int] = None,
        evict_after_last_access: float = 60.0,
        evict_after_creation: float = 0.0,
        ignore: set[str] = set(),
        cache_none: bool = True,
        call_on_delete: Optional[Callable[[Any], None]] = None,
        **kwargs,
    ) -> object:
        """
        Attempt to find a value in the cache.

        If it is not found, the function is called with kwargs as arguments, to produce the value.
        The value is cached.

        all kwargs are prepended to the key

        :param evict_after_last_access: This cache item will be considered stale this number of seconds after
            it was last accessed.
        :param evict_after_creation: This cache item will be considered stale this number of seconds after
            entering the cache.

        """

        def _get_retention_policy(
            for_version: Optional[bool],
            timeout: Optional[int],
            evict_after_last_access: float,
            evict_after_creation: float,
        ) -> tuple[float, float]:
            """
            This method is a backwards compatibility layer to compute a "new-style" retention policy (i.e. that is using
            evict_after_last_access and/or evict_after_creation semantics) from the parameters of the `get_or_else` method.

            :param for_version: Compatibility rules for this deprecated parameter:
                If passed and True, entries will expire <evict_after_last_access>s after their last access (60s default)
                If passed and False, entries will expire a certain number of seconds after entering the cache. This time
                is either <evict_after_creation>, <timeout> or a default of 5000s, whichever is set, inspected in this order.
            :param timeout: Compatibility rules for this deprecated parameter:
                If `for_version=False` and the new-style parameter `evict_after_creation` is not set, this parameter
                controls the expiry time (in seconds) of entries after entering the cache.
                If `for_version` is not set, this parameter is an alias for the new-style parameter `evict_after_creation`.
                (If both are set, the new-style parameter `evict_after_creation` has precedence)
            :param evict_after_last_access: This cache item will be considered stale this number of seconds after
                it was last accessed.
            :param evict_after_creation: This cache item will be considered stale this number of seconds after
                entering the cache.
            """
            _evict_after_last_access: float
            _evict_after_creation: float

            # Legacy `for_version` parameter is used, compute
            # evict_after_last_access and evict_after_creation
            if for_version is not None:
                if for_version:
                    _evict_after_creation = 0.0
                    _evict_after_last_access = evict_after_last_access if evict_after_last_access > 0 else 60.0
                else:
                    _evict_after_last_access = 0.0
                    if evict_after_creation > 0:
                        _evict_after_creation = evict_after_creation
                    elif timeout and timeout > 0:
                        _evict_after_creation = timeout
                    else:
                        _evict_after_creation = 5000.0

            else:
                _evict_after_last_access = evict_after_last_access
                _evict_after_creation = evict_after_creation

                # Set default retention policy if both params are <= 0
                if _evict_after_creation <= 0 and _evict_after_last_access <= 0:
                    if timeout and timeout > 0:
                        # Use legacy parameter timeout if it is set.
                        _evict_after_creation = timeout
                    else:
                        # keep entries alive in the cache for 60s after their last usage by default.
                        _evict_after_last_access = 60.0

            return _evict_after_last_access, _evict_after_creation

        acceptable = {"resource"}
        args = {k: v for k, v in kwargs.items() if k in acceptable and k not in ignore}
        others = sorted([k for k in kwargs.keys() if k not in acceptable and k not in ignore])
        for k in others:
            key = f"{k},{repr(kwargs[k])}" + key
        try:
            return self.find(key, **args)
        except KeyError:
            with self.addLock:
                if key in self.addLocks:
                    lock = self.addLocks[key]
                else:
                    lock = Lock()
                    self.addLocks[key] = lock
            with lock:
                try:
                    value = self.find(key, **args)
                except KeyError:
                    value = function(**kwargs)
                    if cache_none or value is not None:
                        _evict_after_last_access, _evict_after_creation = _get_retention_policy(
                            for_version=for_version,
                            timeout=timeout,
                            evict_after_last_access=evict_after_last_access,
                            evict_after_creation=evict_after_creation,
                        )
                        self.cache_value(
                            key=key,
                            value=value,
                            call_on_delete=call_on_delete,
                            evict_after_last_access=_evict_after_last_access,
                            evict_after_creation=_evict_after_creation,
                            **args,
                        )
            with self.addLock:
                del self.addLocks[key]
            return value

    def __enter__(self) -> None:
        """
        Assumed to be called under activity_lock.
        Clean stale entries before using the cache.
        """
        self.clean_stale_entries()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """
        When done using the cache, reset the expiry time
        of all cache items that should be refreshed after
        access.
        """
        self.touch_used_cache_items()



"""
Copyright 2019 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import os
import random
import socket

import psutil
import pytest
from tornado import netutil

import inmanta.agent.config as cfg
from inmanta import protocol
from inmanta.config import Config, Option, option_as_default
from inmanta.const import ClientType
from inmanta.server.protocol import Server, ServerSlice


def test_options(monkeypatch):
    """
    Test basic functionalities of the configuration framework:
        - setting/getting option values
        - overriding using an environment variable
        - using the get_environment_variable method to return associated env variable
    """
    configa = Option("test", "a", "markerA", "test a docs")
    configb = Option("test", "B", option_as_default(configa), "test b docs")
    configc = Option("test", "c", "defaultc", "docstringc")
    configd = Option("test.section", "d.option", "defaultd", "docstringd")

    monkeypatch.setenv("INMANTA_TEST_C", "environ_c")
    monkeypatch.setenv("INMANTA_TEST_SECTION_D_OPTION", "environ_d")

    assert "test.a" in configb.get_default_desc()

    Config.load_config()

    assert configb.get() == "markerA"
    configa.set("MA2")
    assert configb.get() == "MA2"
    configb.set("MB2")
    assert configb.get() == "MB2"
    assert configc.get() == "environ_c"
    assert configd.get() == "environ_d"

    option_to_env_var_name = {
        configa: "INMANTA_TEST_A",
        configb: "INMANTA_TEST_B",
        configc: "INMANTA_TEST_C",
        configd: "INMANTA_TEST_SECTION_D_OPTION",
    }
    for option, expected_env_var_name in option_to_env_var_name.items():
        assert option.get_environment_variable() == expected_env_var_name


def test_configfile_hierarchy(monkeypatch, tmpdir):
    """
    Test the hierarchy when a config option is set in multiple places:

        - in the default main config file /etc/inmanta/inmanta.cfg
        - in a config directory --config-dir cli option
        - in dot files (.inmanta or .inmanta.cfg)
        - in a config file passed via --config cli option
        - via environment variable

        Lower in the list means higher precedence and ties are broken
        by alphabetical ordering of file names.
    """
    etc_inmanta_dir = os.path.join(tmpdir, "etc", "inmanta")
    os.makedirs(etc_inmanta_dir, exist_ok=False)

    main_inmanta_cfg_file = os.path.join(etc_inmanta_dir, "inmanta.cfg")

    inmanta_d_dir = os.path.join(etc_inmanta_dir, "inmanta.d")
    os.mkdir(inmanta_d_dir)

    inmanta_d_cfg_file01 = os.path.join(inmanta_d_dir, "01-dbconfig.cfg")
    inmanta_d_cfg_file02 = os.path.join(inmanta_d_dir, "02-dbconfig.cfg")
    inmanta_d_cfg_file_no_cfg_extension = os.path.join(inmanta_d_dir, "03-config")

    dot_inmanta_file = os.path.join(tmpdir, ".inmanta")
    dot_inmanta_cfg_file = os.path.join(tmpdir, ".inmanta.cfg")

    min_c_file = os.path.join(tmpdir, "custom.cfg")

    monkeypatch.setenv("INMANTA_SERVER_AUTH", "true")
    monkeypatch.setenv("INMANTA_SERVER_AGENT_TIMEOUT", "60")

    with open(main_inmanta_cfg_file, "w", encoding="utf-8") as f:
        f.write(
            """
[server]
auth=false
db-connection-pool-max-size=2
[config]
log-dir=/log
[database]
host=host1
name=db1
port=1234
username=non-default-name-0
[influxdb]
host=host1
interval=10
tags=tag1=value1
        """
        )

    with open(inmanta_d_cfg_file01, "w", encoding="utf-8") as f:
        f.write(
            """
[database]
host=host2
name=db2
[influxdb]
host=host2
        """
        )

    with open(inmanta_d_cfg_file02, "w", encoding="utf-8") as f:
        f.write(
            """
[database]
port=5678
[influxdb]
host=host3
interval=20
        """
        )

    with open(inmanta_d_cfg_file_no_cfg_extension, "w", encoding="utf-8") as f:
        f.write(
            """
[database]
port=9999
        """
        )

    with open(dot_inmanta_file, "w", encoding="utf-8") as f:
        f.write(
            """
[database]
host=host3
username=non-default-name-1
[influxdb]
tags=tag2=value2
        """
        )

    with open(dot_inmanta_cfg_file, "w", encoding="utf-8") as f:
        f.write(
            """
[database]
username=non-default-name-2
[server]
db-connection-pool-max-size=3
        """
        )

    with open(min_c_file, "w", encoding="utf-8") as f:
        f.write(
            """
[server]
db-connection-pool-max-size=5
        """
        )

    os.chdir(tmpdir)
    Config.load_config(min_c_config_file=min_c_file, config_dir=inmanta_d_dir, main_cfg_file=main_inmanta_cfg_file)

    assert Config.get("config", "log-dir") == "/log"
    assert Config.get("database", "host") == "host3"
    assert Config.get("database", "name") == "db2"
    assert Config.get("database", "port") == 5678
    assert Config.get("influxdb", "host") == "host3"
    assert Config.get("influxdb", "interval") == 20
    assert Config.get("influxdb", "tags")["tag2"] == "value2"
    assert Config.get("database", "username") == "non-default-name-2"
    assert Config.get("server", "db_connection_pool_max_size") == 5
    assert Config.get("server", "auth")
    assert Config.get("server", "agent-timeout") == 60


async def test_bind_address_ipv4(async_finalizer):
    """This test case check if the Inmanta server doesn't bind on another interface than 127.0.0.1 when bind-address is equal
    to 127.0.0.1. Procedure:
        1) Get free port on all interfaces.
        2) Bind that port on a non-loopback interface, so it's not available for the inmanta server anymore.
        3) Start the Inmanta server with bind-address 127.0.0.1. and execute an API call
    """

    @protocol.method(path="/test", operation="POST", client_types=[ClientType.api])
    async def test_endpoint():
        pass

    class TestSlice(ServerSlice):
        @protocol.handle(test_endpoint)
        async def test_endpoint_handle(self):
            return 200

    # Select a bind address which is not on the loopback interface
    non_loopback_interfaces = [
        address
        for (name, adresses) in psutil.net_if_addrs().items()
        for (family, address, netmask, ptp, broadcast) in adresses
        if name != "lo" and family == socket.AF_INET
    ]
    bind_addr = random.choice(non_loopback_interfaces)

    # Get free port on all interfaces
    sock = netutil.bind_sockets(0, "0.0.0.0", family=socket.AF_INET)[0]
    _addr, free_port = sock.getsockname()
    sock.close()

    # Bind port on non-loopback interface
    sock = netutil.bind_sockets(free_port, bind_addr, family=socket.AF_INET)[0]
    try:
        # Configure server
        Config.load_config()
        Config.set("server", "bind-port", str(free_port))
        Config.set("server", "bind-address", "127.0.0.1")
        Config.set("client_rest_transport", "port", str(free_port))

        # Start server
        rs = Server()
        rs.add_slice(TestSlice("test"))
        await rs.start()
        async_finalizer(rs.stop)

        # Check if server is reachable on loopback interface
        client = protocol.Client("client")
        result = await client.test_endpoint()
        assert result.code == 200
    finally:
        sock.close()


async def test_bind_address_ipv6(async_finalizer) -> None:
    @protocol.method(path="/test", operation="POST", client_types=[ClientType.api])
    async def test_endpoint():
        pass

    class TestSlice(ServerSlice):
        @protocol.handle(test_endpoint)
        async def test_endpoint_handle(self):
            return 200

    # Get free port on all interfaces
    sock = netutil.bind_sockets(0, "::", family=socket.AF_INET6)[0]
    (_addr, free_port, _flowinfo, _scopeid) = sock.getsockname()
    sock.close()

    # Configure server
    Config.load_config()
    Config.set("server", "bind-port", str(free_port))
    Config.set("server", "bind-address", "::1")
    Config.set("client_rest_transport", "port", str(free_port))
    Config.set("client_rest_transport", "host", "::1")

    # Start server
    rs = Server()
    rs.add_slice(TestSlice("test"))
    await rs.start()
    async_finalizer(rs.stop)

    client = protocol.Client("client")
    # Check if server is reachable on loopback interface
    result = await client.test_endpoint()
    assert result.code == 200


async def test_bind_port(unused_tcp_port, async_finalizer):
    @protocol.method(path="/test", operation="POST", client_types=[ClientType.api])
    async def test_endpoint():
        pass

    class TestSlice(ServerSlice):
        @protocol.handle(test_endpoint)
        async def test_endpoint_handle(self):
            return 200

    async def assert_port_bound():
        # Start server
        rs = Server()
        rs.add_slice(TestSlice("test"))
        await rs.start()
        async_finalizer(rs.stop)

        # Check if server is reachable on loopback interface
        client = protocol.Client("client")
        result = await client.test_endpoint()
        assert result.code == 200
        await rs.stop()

    Config.load_config()
    Config.set("server", "bind-port", str(unused_tcp_port))
    Config.set("client_rest_transport", "port", str(unused_tcp_port))
    await assert_port_bound()


def test_option_is_list():
    option: Option = Option("test", "list", "default,values", "documentation", cfg.is_list)
    option.set("some,values")
    assert option.get() == ["some", "values"]


def test_option_is_list_empty():
    option: Option = Option("test", "list", "default,values", "documentation", cfg.is_list)
    option.set("")
    assert option.get() == []


def test_option_is_lower_bounded_int():
    lower_bound = 1
    option: Option = Option("test", "lb_int", lower_bound, "documentation", cfg.is_lower_bounded_int(lower_bound))
    option.set("2")
    assert option.get() == 2

    option.set("0")
    with pytest.raises(ValueError) as exc_info:
        option.get()

    assert f"Value can not be lower than {lower_bound}" in str(exc_info.value)



"""
Copyright 2024 Inmanta
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Contact: code@inmanta.com
"""

import asyncio
import datetime
import logging
import typing
import uuid
from asyncio import InvalidStateError, Lock
from collections import defaultdict
from collections.abc import Mapping
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Any, Optional

import inmanta.agent.cache
import inmanta.loader as loader
import inmanta.protocol
import inmanta.util
from inmanta import const, data, env, tracing
from inmanta.agent import executor, handler
from inmanta.agent.executor import DeployReport, DryrunReport, FailedInmantaModules, GetFactReport, ResourceDetails
from inmanta.agent.handler import HandlerAPI, SkipResource, SkipResourceForDependencies
from inmanta.const import NAME_RESOURCE_ACTION_LOGGER, ParameterSource
from inmanta.data.model import AttributeStateChange
from inmanta.resources import Resource
from inmanta.types import ResourceIdStr, ResourceVersionIdStr
from inmanta.util import NamedLock, join_threadpools


class InProcessExecutor(executor.Executor, executor.AgentInstance):
    """
    This is an executor that executes in the process it is started in. It executes the appropriate handler code in this same
    process.

    !!! This executor takes no steps to prevent handlers to mutate the resources passed into the methods !!!
    """

    def __init__(
        self,
        agent_name: str,
        agent_uri: str,
        environment: uuid.UUID,
        client: inmanta.protocol.SessionClient,
        eventloop: asyncio.AbstractEventLoop,
        parent_logger: logging.Logger,
    ):
        self.name = agent_name
        self.client = client
        self.uri = agent_uri

        # For agentinstance api
        self.eventloop = eventloop
        self.sessionid = client._sid
        self.environment = environment

        # threads to work
        self.thread_pool: ThreadPoolExecutor = ThreadPoolExecutor(1, thread_name_prefix="Pool_%s" % self.name)

        self._cache = inmanta.agent.cache.AgentCache(self)
        # This lock ensures cache entries can not be cleaned up when
        # the executor is actively working and vice versa
        self.activity_lock = asyncio.Lock()

        self.logger: logging.Logger = parent_logger.getChild(self.name)
        self.resource_action_logger = logging.getLogger(NAME_RESOURCE_ACTION_LOGGER).getChild(self.name)

        self._stopped = False

        self.failed_modules: FailedInmantaModules = dict()

        self.cache_cleanup_tick_rate = inmanta.agent.config.agent_cache_cleanup_tick_rate.get()
        self.periodic_cache_cleanup_job: Optional[asyncio.Task[None]] = None

    async def start(self) -> None:
        self.periodic_cache_cleanup_job = asyncio.create_task(self.cleanup_stale_cache_entries())

    async def cleanup_stale_cache_entries(self) -> None:
        """
        Periodically cleans up stale entries in the cache. The clean_stale_entries
        has to be called on the thread pool because it might call finalizers.
        """
        reschedule_interval: int = self.cache_cleanup_tick_rate
        while not self._stopped:
            async with self.activity_lock:
                if self._stopped:
                    return
                try:
                    await asyncio.get_running_loop().run_in_executor(self.thread_pool, self._cache.clean_stale_entries)
                except Exception:
                    # Make sure we don't drop out of the while loop if an exception occurs.
                    self.logger.exception(
                        "An unexpected exception occurred while cleaning up the cache of agent %s.", self.name
                    )
            await asyncio.sleep(reschedule_interval)

    async def stop(self) -> None:
        if self._stopped:
            return
        self._stopped = True
        if self.periodic_cache_cleanup_job:
            try:
                self.periodic_cache_cleanup_job.cancel()
                await self.periodic_cache_cleanup_job
            except asyncio.CancelledError:
                pass
        async with self.activity_lock:
            await asyncio.get_running_loop().run_in_executor(self.thread_pool, self._cache.close)
        self.thread_pool.shutdown(wait=False)

    async def join(self) -> None:
        """
        Called after stop to ensure complete shutdown

        :param thread_pool_finalizer: all threadpools that should be joined should be added here.
        """
        assert self._stopped
        await join_threadpools([self.thread_pool])
        if self.periodic_cache_cleanup_job:
            try:
                await self.periodic_cache_cleanup_job
            except asyncio.CancelledError:
                self.periodic_cache_cleanup_job = None

    def is_stopped(self) -> bool:
        return self._stopped

    async def get_provider(self, resource: Resource) -> HandlerAPI[Any]:
        provider = handler.Commander.get_provider(agent=self, resource=resource)
        provider.set_cache(self._cache)
        return provider

    def _log_deserialization_error(self, resource_details: ResourceDetails, cause: Exception) -> data.LogLine:
        msg = data.LogLine.log(
            level=const.LogLevel.ERROR,
            msg="Unable to deserialize %(resource_id)s: %(cause)s",
            resource_id=resource_details.rvid,
            cause=cause,
            timestamp=datetime.datetime.now().astimezone(),
        )
        msg.write_to_logger_for_resource(
            agent=resource_details.id.agent_name, resource_version_string=resource_details.rvid, exc_info=True
        )
        return msg

    async def _execute(
        self,
        resource: Resource,
        ctx: handler.HandlerContext,
        requires: Mapping[ResourceIdStr, const.ResourceState],
    ) -> None:
        """
        Get the handler for a given resource and run its ``deploy`` method.

        :param resource: The resource to deploy.
        :param ctx: The context to use during execution of this deploy.
        :param requires: A dictionary that maps each dependency of the resource to be deployed, to its latest resource
                         state that was not `deploying'.
        """
        # setup provider

        provider: Optional[HandlerAPI[Any]] = None
        try:
            provider = await self.get_provider(resource)
        except Exception:
            ctx.set_resource_state(const.HandlerResourceState.unavailable)
            ctx.exception("Unable to find a handler for %(resource_id)s", resource_id=resource.id.resource_version_str())
            return
        else:
            # main execution
            try:

                def resolve_and_deploy(
                    provider: HandlerAPI[Resource],
                    ctx: handler.HandlerContext,
                    resource: Resource,
                    requires: Mapping[ResourceIdStr, const.ResourceState],
                ) -> None:
                    resource.resolve_all_references(ctx)
                    provider.deploy(ctx, resource, requires)

                await asyncio.get_running_loop().run_in_executor(
                    self.thread_pool,
                    resolve_and_deploy,
                    provider,
                    ctx,
                    resource,
                    requires,
                )
                if ctx.status is None:
                    ctx.set_resource_state(const.HandlerResourceState.deployed)
            except SkipResourceForDependencies as e:
                ctx.set_resource_state(const.HandlerResourceState.skipped_for_dependency)
                ctx.warning(
                    msg="Resource %(resource_id)s was skipped: %(reason)s",
                    resource_id=resource.id,
                    reason=e.args,
                )
            except SkipResource as e:
                ctx.set_resource_state(const.HandlerResourceState.skipped)
                ctx.warning(msg="Resource %(resource_id)s was skipped: %(reason)s", resource_id=resource.id, reason=e.args)
            except Exception as e:
                ctx.set_resource_state(const.HandlerResourceState.failed)
                ctx.exception(
                    "An error occurred during deployment of %(resource_id)s (exception: %(exception)s",
                    resource_id=str(resource.id),
                    exception=repr(e),
                )
        finally:
            if provider is not None:
                provider.close()

    @tracing.instrument("InProcessExecutor.execute", extract_args=True)
    async def execute(
        self,
        action_id: uuid.UUID,
        gid: uuid.UUID,
        resource_details: ResourceDetails,
        reason: str,
        requires: Mapping[ResourceIdStr, const.ResourceState],
    ) -> DeployReport:
        try:
            resource: Resource = Resource.deserialize(resource_details.attributes)
        except Exception as e:
            msg = self._log_deserialization_error(resource_details, e)
            return DeployReport.undeployable(resource_details.rvid, action_id, msg)

        ctx = handler.HandlerContext(resource, action_id=action_id, logger=self.resource_action_logger)

        ctx.debug(
            "Start run because %(reason)s.",
            reason=reason,
            deploy_id=gid,
            resource=resource_details.id,
        )

        async with self.activity_lock:
            with self._cache:
                await self._execute(resource, ctx=ctx, requires=requires)

        ctx.debug(
            "End run for resource %(r_id)s in deploy %(deploy_id)s",
            r_id=resource_details.rvid,
            deploy_id=gid,
        )

        if ctx.facts:
            ctx.debug("Sending facts to the server")
            set_fact_response = await self.client.set_parameters(tid=self.environment, parameters=ctx.facts)
            if set_fact_response.code != 200:
                ctx.error("Failed to send facts to the server %s", set_fact_response.result)

        return DeployReport.from_ctx(resource_details.rvid, ctx)

    async def dry_run(
        self,
        resource: ResourceDetails,
        dry_run_id: uuid.UUID,
    ) -> DryrunReport:
        """
        Perform a dryrun for the given resources

        :param resource: Resource for which to perform a dryrun.
        :param dry_run_id: id for this dryrun
        """
        async with self.activity_lock:
            with self._cache:
                started = datetime.datetime.now().astimezone()
                try:
                    resource_obj: Resource = Resource.deserialize(resource.attributes)
                except Exception as e:
                    msg = self._log_deserialization_error(resource, e)
                    return DryrunReport(
                        rvid=resource.rvid,
                        dryrun_id=dry_run_id,
                        changes={"handler": AttributeStateChange(current="FAILED", desired="Resource Deserialization Failed")},
                        started=started,
                        finished=datetime.datetime.now().astimezone(),
                        messages=[msg],
                        resource_state=const.ResourceState.unavailable,
                    )
                assert resource_obj is not None
                ctx = handler.HandlerContext(resource_obj, True, logger=self.resource_action_logger)
                provider = None

                dryrun_result: Optional[DryrunReport] = None
                resource_id: ResourceVersionIdStr = resource.rvid

                try:
                    self.resource_action_logger.debug("Running dryrun for %s", resource_id)

                    try:
                        provider = await self.get_provider(resource_obj)
                    except Exception as e:
                        ctx.exception(
                            "Unable to find a handler for %(resource_id)s (exception: %(exception)s",
                            resource_id=resource_id,
                            exception=str(e),
                        )
                        dryrun_result = DryrunReport(
                            rvid=resource_id,
                            dryrun_id=dry_run_id,
                            changes={"handler": AttributeStateChange(current="FAILED", desired="Unable to find a handler")},
                            started=started,
                            finished=datetime.datetime.now().astimezone(),
                            messages=ctx.logs,
                        )
                    else:
                        try:

                            def resolve_and_dryrun(
                                provider: HandlerAPI[Resource],
                                ctx: handler.HandlerContext,
                                resource: Resource,
                            ) -> None:
                                resource.resolve_all_references(ctx)
                                provider.execute(ctx, resource, True)

                            await asyncio.get_running_loop().run_in_executor(
                                self.thread_pool, resolve_and_dryrun, provider, ctx, resource_obj
                            )

                            changes = ctx.changes
                            if changes is None:
                                changes = {}
                            if ctx.status == const.ResourceState.failed:
                                changes["handler"] = AttributeStateChange(current="FAILED", desired="Handler failed")
                            dryrun_result = DryrunReport(
                                rvid=resource_id,
                                dryrun_id=dry_run_id,
                                changes=changes,
                                started=started,
                                finished=datetime.datetime.now().astimezone(),
                                messages=ctx.logs,
                            )
                        except Exception as e:
                            ctx.exception(
                                "Exception during dryrun for %(resource_id)s (exception: %(exception)s",
                                resource_id=resource.rvid,
                                exception=str(e),
                            )
                            changes = ctx.changes
                            if changes is None:
                                changes = {}
                            changes["handler"] = AttributeStateChange(current="FAILED", desired="Handler failed")
                            dryrun_result = DryrunReport(
                                rvid=resource_id,
                                dryrun_id=dry_run_id,
                                changes=changes,
                                started=started,
                                finished=datetime.datetime.now().astimezone(),
                                messages=ctx.logs,
                            )

                except Exception as e:
                    ctx.exception(
                        "Unable to process resource %(resource_id)s for dryrun (exception: %(exception)s",
                        resource_id=resource.rvid,
                        exception=str(e),
                    )
                    dryrun_result = DryrunReport(
                        rvid=resource_id,
                        dryrun_id=dry_run_id,
                        changes={"handler": AttributeStateChange(current="FAILED", desired="Resource Deserialization Failed")},
                        started=started,
                        finished=datetime.datetime.now().astimezone(),
                        messages=ctx.logs,
                    )
                finally:
                    if provider is not None:
                        provider.close()
                    assert dryrun_result is not None, "Dryrun result cannot be None"
                    return dryrun_result

    async def get_facts(self, resource: ResourceDetails) -> GetFactReport:
        """
        Get facts for a given resource
        :param resource: The resource for which to get facts.
        """
        provider = None
        started = datetime.datetime.now().astimezone()
        try:
            try:
                resource_obj: Resource = Resource.deserialize(resource.attributes)
            except Exception as e:
                msg = self._log_deserialization_error(resource, e)
                return GetFactReport(
                    resource_id=resource.rvid,
                    action_id=None,
                    parameters=[],
                    started=started,
                    finished=datetime.datetime.now().astimezone(),
                    messages=[msg],
                    success=False,
                    error_msg=f"Unable to deserialize resource {resource.id}",
                    resource_state=const.ResourceState.unavailable,
                )
            assert resource_obj is not None
            ctx = handler.HandlerContext(resource_obj, logger=self.resource_action_logger)
            async with self.activity_lock:
                try:
                    with self._cache:
                        provider = await self.get_provider(resource_obj)

                        def resolve_and_getfact(
                            provider: HandlerAPI[Resource],
                            ctx: handler.HandlerContext,
                            resource: Resource,
                        ) -> dict[str, str]:
                            resource.resolve_all_references(ctx)
                            return provider.check_facts(ctx, resource)

                        result = await asyncio.get_running_loop().run_in_executor(
                            self.thread_pool, resolve_and_getfact, provider, ctx, resource_obj
                        )

                        parameters = [
                            {
                                "id": name,
                                "value": value,
                                "resource_id": resource.rid,
                                "source": ParameterSource.fact.value,
                            }
                            for name, value in result.items()
                        ]
                        # Add facts set via the set_fact() method of the HandlerContext
                        parameters.extend(ctx.facts)

                    return GetFactReport(
                        resource_id=resource.rvid,
                        action_id=ctx.action_id,
                        parameters=parameters,
                        started=started,
                        finished=datetime.datetime.now().astimezone(),
                        messages=ctx.logs,
                        success=True,
                    )

                except Exception:
                    error_msg = "Unable to retrieve facts for resource %s" % resource.id
                    self.resource_action_logger.exception(error_msg)
                    return GetFactReport(
                        resource_id=resource.rvid,
                        action_id=ctx.action_id,
                        parameters=[],
                        started=started,
                        finished=datetime.datetime.now().astimezone(),
                        messages=ctx.logs,
                        success=False,
                        error_msg=error_msg,
                    )

        except Exception:
            error_msg = "Unable to find a handler for %s" % resource.id
            self.resource_action_logger.exception(error_msg)
            return GetFactReport(
                resource_id=resource.rvid,
                action_id=None,
                parameters=[],
                started=started,
                finished=datetime.datetime.now().astimezone(),
                messages=[],
                success=False,
                error_msg=error_msg,
            )
        finally:
            if provider is not None:
                provider.close()


class InProcessExecutorManager(executor.ExecutorManager[InProcessExecutor]):
    """
    This is the executor that provides the backward compatible behavior, conforming to the agent in ISO7.
    It is no longer used outside of testing.

    It spawns an InProcessExecutor and makes sure all code is installed and loadable locally.
    """

    def __init__(
        self,
        environment: uuid.UUID,
        client: inmanta.protocol.SessionClient,
        eventloop: asyncio.AbstractEventLoop,
        parent_logger: logging.Logger,
        thread_pool: ThreadPoolExecutor,
        code_dir: str,
        env_dir: str,
        code_loader: bool = True,
    ) -> None:
        self.environment = environment
        self.client = client
        self.eventloop = eventloop
        self.logger = parent_logger
        self.thread_pool = thread_pool

        self.executors: dict[str, InProcessExecutor] = {}
        self._creation_locks: inmanta.util.NamedLock = inmanta.util.NamedLock()

        self._loader: loader.CodeLoader | None = None
        self._env: env.VirtualEnv | None = None
        self._running = False

        if code_loader:
            self._env = env.VirtualEnv(env_dir)
            self._env.use_virtual_env()
            self._loader = loader.CodeLoader(code_dir, clean=True)
            # Lock to ensure only one actual install runs at a time
            self._loader_lock: asyncio.Lock = Lock()
            # Keep track for each resource type of the last loaded version
            self._last_loaded_version: dict[str, executor.ExecutorBlueprint | None] = defaultdict(lambda: None)
            # Per-resource lock to serialize all actions per resource
            self._resource_loader_lock: NamedLock = NamedLock()

    async def stop(self) -> None:
        self._running = False
        for child in self.executors.values():
            await child.stop()

    async def start(self) -> None:
        assert all(e.is_stopped() for e in self.executors.values())
        await self.join(thread_pool_finalizer=[], timeout=const.SHUTDOWN_GRACE_IOLOOP * 0.9)
        self.executors.clear()
        self._running = True

    def get_environment_manager(self) -> None:
        return None

    async def stop_all_executors(self) -> list[InProcessExecutor]:
        raise NotImplementedError("Not used")

    async def stop_for_agent(self, agent_name: str) -> list[InProcessExecutor]:
        if agent_name in self.executors:
            out = self.executors[agent_name]
            del self.executors[agent_name]
            await out.stop()
        return []

    async def join(self, thread_pool_finalizer: list[ThreadPoolExecutor], timeout: float) -> None:
        assert not self._running
        await asyncio.gather(*(child.join() for child in self.executors.values()))

    async def get_executor(
        self, agent_name: str, agent_uri: str, code: typing.Collection[executor.ResourceInstallSpec]
    ) -> InProcessExecutor:
        """
        Retrieves an Executor for a given agent with the relevant handler code loaded in its venv.
        If an Executor does not exist for the given configuration, a new one is created.

        :param agent_name: The name of the agent for which an Executor is being retrieved or created.
        :param agent_uri: The name of the host on which the agent is running.
        :param code: Collection of ResourceInstallSpec defining the configuration for the Executor i.e.
            which resource types it can act on and all necessary information to install the relevant
            handler code in its venv. Must have at least one element.
        :return: An Executor instance
        """
        if not self._running:
            raise InvalidStateError("This executor manager is not running")
        if not code:
            raise ValueError(f"{self.__class__.__name__}.get_executor() expects at least one resource install specification")
        if agent_name in self.executors:
            out = self.executors[agent_name]
        else:
            async with self._creation_locks.get(agent_name):
                if agent_name in self.executors:
                    out = self.executors[agent_name]
                else:
                    out = InProcessExecutor(agent_name, agent_uri, self.environment, self.client, self.eventloop, self.logger)
                    await out.start()
                    self.executors[agent_name] = out
        assert out.uri == agent_uri
        out.failed_modules = await self.ensure_code(code)

        return out

    async def ensure_code(self, code: typing.Collection[executor.ResourceInstallSpec]) -> executor.FailedInmantaModules:
        """Ensure that the code for the given environment and version is loaded"""

        failed_to_load: FailedInmantaModules = defaultdict(dict)

        if self._loader is None:
            return failed_to_load

        for resource_install_spec in code:
            # only one logical thread can load a particular resource type at any time
            async with self._resource_loader_lock.get(resource_install_spec.resource_type):
                # stop if the last successful load was this one
                # The combination of the lock and this check causes the reloads to naturally 'batch up'
                if self._last_loaded_version[resource_install_spec.resource_type] == resource_install_spec.blueprint:
                    self.logger.debug(
                        "Handler code already installed for %s version=%d",
                        resource_install_spec.resource_type,
                        resource_install_spec.model_version,
                    )
                    continue

                try:
                    # Install required python packages and the list of ``ModuleSource`` with the provided pip config
                    self.logger.debug(
                        "Installing handler %s version=%d",
                        resource_install_spec.resource_type,
                        resource_install_spec.model_version,
                    )
                    await self._install(resource_install_spec.blueprint)
                    self.logger.debug(
                        "Installed handler %s version=%d",
                        resource_install_spec.resource_type,
                        resource_install_spec.model_version,
                    )

                    self._last_loaded_version[resource_install_spec.resource_type] = resource_install_spec.blueprint
                except Exception as e:
                    self.logger.exception(
                        "Failed to install handler %s version=%d",
                        resource_install_spec.resource_type,
                        resource_install_spec.model_version,
                    )

                    if resource_install_spec.resource_type not in failed_to_load:
                        failed_to_load[resource_install_spec.resource_type][resource_install_spec.resource_type] = Exception(
                            f"Failed to install handler {resource_install_spec.resource_type} "
                            f"version={resource_install_spec.model_version}: {e}"
                        ).with_traceback(e.__traceback__)
                    self._last_loaded_version[resource_install_spec.resource_type] = None

        return failed_to_load

    async def _install(self, blueprint: executor.ExecutorBlueprint) -> None:
        if self._env is None or self._loader is None:
            raise Exception("Unable to load code when agent is started with code loading disabled.")

        async with self._loader_lock:
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(
                self.thread_pool,
                self._env.install_for_config,
                inmanta.util.parse_requirements(blueprint.requirements),
                blueprint.pip_config,
            )
            await loop.run_in_executor(self.thread_pool, self._loader.deploy_version, blueprint.sources)



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    schema = """
    -- Add the 'soft_delete' column to support marking a resource set for deletion
    -- and deleting it iff the model doesn't export resources for it.

    ALTER TABLE public.compile
    ADD COLUMN soft_delete boolean DEFAULT false NOT NULL;
    """
    await connection.execute(schema)



"""
Copyright 2022 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import inmanta.types
from utils import get_resource

"""
This module contains tests related to database concurrency issues. Whenever we fix a concurrency issue, be it performance
or deadlock related, a test should be added to ensure this occurence can not accidentally be introduced again.
"""

import asyncio
import datetime
import uuid
from collections import abc
from typing import Optional, TypeVar

import asyncpg
import pytest

from inmanta import const, data
from inmanta.protocol.common import Result


def slowdown_queries(
    monkeypatch,
    *,
    cls: type[data.BaseDocument] = data.BaseDocument,
    query_funcs: Optional[abc.Collection[str]] = None,
    delay: float = 1,
) -> None:
    """
    Introduces an artificial delay after each query execution through data.Document in order to increase the likelyhood of
    concurrent transactions.

    :param cls: The class to slow down the methods for.
    :param query_funcs: The names of the methods to add a delay to.
    :param delay: The amount of seconds to delay.
    """
    query_funcs = (
        query_funcs
        if query_funcs
        else [
            "select_query",
            "_fetchval",
            "_fetch_int",
            "_fetchrow",
            "_fetch_query",
            "_execute_query",
            "insert_many",
        ]
    )

    F = TypeVar("F", bound=abc.Coroutine)

    def patch_method(method: F) -> F:
        clsmethod: bool = hasattr(method, "__func__")
        func: F = method.__func__ if clsmethod else method

        async def patched(*args, **kwargs) -> object:
            # call unbound original method with cls object bound to the patched method
            result: object = await func(*args, **kwargs)
            await asyncio.sleep(delay)
            return result

        return classmethod(patched) if clsmethod else patched

    for query_func in query_funcs:
        monkeypatch.setattr(cls, query_func, patch_method(getattr(cls, query_func)))


@pytest.mark.slowtest
async def test_4889_deadlock_delete_resource_action_insert(monkeypatch, environment: str) -> None:
    """
    Verify that no deadlock exists between the delete of a version and the insert of a ResourceAction for that same version.
    """
    env_id: uuid.UUID = uuid.UUID(environment)

    version: int = 1
    confmodel: data.ConfigurationModel = data.ConfigurationModel(
        environment=env_id,
        version=version,
        date=datetime.datetime.now().astimezone(),
        total=1,
        version_info={},
        is_suitable_for_partial_compiles=False,
    )
    await confmodel.insert()

    resource = inmanta.types.ResourceVersionIdStr(f"mymod::myresource[myagent,id=1],v={version}")
    await data.Resource.new(
        environment=env_id,
        status=const.ResourceState.available,
        resource_version_id=resource,
        attributes={},
    ).insert()

    # artificially slow down ResourceAction queries to increase deadlock probability
    slowdown_queries(monkeypatch, cls=data.ResourceAction, delay=1)

    insert: abc.Awaitable[None] = data.ResourceAction(
        environment=env_id,
        version=version,
        resource_version_ids=[resource],
        action_id=uuid.uuid4(),
        action=const.ResourceAction.deploy,
        started=datetime.datetime.now(),
    ).insert()

    async def delete() -> None:
        # Make sure insert starts first so it can acquire its first lock.
        await asyncio.sleep(0.5)
        await confmodel.delete_cascade()

    # Verify that this does not raise a deadlock exception. A failure on the insert is expected and acceptable if the delete
    # wins the race.
    try:
        await asyncio.gather(insert, delete())
    except asyncpg.ForeignKeyViolationError:
        pass


@pytest.mark.slowtest
@pytest.mark.parametrize("no_agent", [True])
async def test_release_version_concurrently(
    monkeypatch, server, client, environment: str, clienthelper, no_agent: bool
) -> None:
    version1 = await clienthelper.get_version()
    resource1 = get_resource(version1, key="test1")
    await clienthelper.put_version_simple(resources=[resource1], version=version1)

    version2 = await clienthelper.get_version()
    resource1 = get_resource(version2, key="test1")
    await clienthelper.put_version_simple(resources=[resource1], version=version2)

    slowdown_queries(monkeypatch)

    async def release() -> Result[None]:
        return await client.release_version(environment, version2)

    f1 = asyncio.create_task(release())
    f2 = asyncio.create_task(release())

    # get results
    r1 = await f1
    r2 = await f2

    # One should have made it, the other was too late
    assert {r1.code, r2.code} == {200, 409}

    # releasing an older version is always too late
    r3 = await client.release_version(environment, version1)
    assert r3.code == 409



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com

This file contains the caching mechanism for the forking executor

The complexity is that:
1. we need to cache processes on blueprint id
2. we need to cache executors (running on processes) on executor id
3. invalidate both caches if a process dies
4. invalidate executors based on idle time
5. invalidate executors based on nr per agent
6. closed processes if they are no longer in use
7. never invalidate executors that have work in flight

The solution is a combination of pool manager/pool members that can:
1. can perform invalidation according to various stategies
2. has fairly accurate membership tracking (for requirements 4  and 5) through events/listeners
    (cache responds correctly to crashes)
3. members can be contained in multiple cache

To make this work, we propagate the 'closed' event, that indicates when a pool member has terminated.
We use it for cache eviction and failure propagation.

The `closing` status is tracked on each member and checked when taking an item out of the cache.
As such, the cache can contain closing items. These are replaced when being checked out.
We don't propagate this as an event, as they will eventually get a closed event.

The intended usage is that:
1. A single Pool Manager manages a pool of processes (no eviction policy)
2. each process is a pool manager for a set of executors (evict self when pool is empty, on receival of closed event)
3. an overarching pool manager also caches executors on top of the other two (i.e. it never creates new things),
    it evicts based on time and executors per agent
"""

import abc
import asyncio
import datetime
import logging
from asyncio import CancelledError, Task
from typing import Any, Callable, Coroutine, Generic, Optional, TypeVar

import inmanta.util
from inmanta.const import LOG_LEVEL_TRACE

LOGGER = logging.getLogger(__name__)

TPoolID = TypeVar("TPoolID")
TIntPoolID = TypeVar("TIntPoolID")


class PoolMember(abc.ABC, Generic[TIntPoolID]):
    """
    Item that can live in a pool

    The lifecycle if this item is:
    - running (after construction)
    - shutting down (after call to request_shutdown)
    - shutdown (after call to set_shutdown)

    Implementors of sub-classes are expected to
    - override request_shutdown to start the shutdown of the underlying resource
    - at the end of request_shutdown, call set_shutdown to indicate we are down

    After request_shutdown is called, this member will be considered invalid.
        When checked out of any pool a replacement will be produced.
    When set_shutdown is called, the shutdown event will be sent to all pools containing this member.
        This will cause it to be evicted
    """

    def __init__(self, my_id: TIntPoolID):
        self.id = my_id

        # Time based expiry
        self._last_used: datetime.datetime = datetime.datetime.now().astimezone()

        # state tracking
        self.shutting_down = False
        self.shut_down = False

        # event propagation
        self.termination_listeners: list[Callable[[PoolMember[TPoolID]], Coroutine[Any, Any, Any]]] = []

    @property
    def running(self) -> bool:
        return not self.shutting_down and not self.shut_down

    @property
    def last_used(self) -> datetime.datetime:
        return self._last_used

    def touch(self) -> None:
        """Update time last used"""
        self._last_used = datetime.datetime.now().astimezone()

    def can_be_cleaned_up(self) -> bool:
        """
        Return true if this member is not inhibited from cleanup
        """
        return True

    def get_idle_time(self) -> datetime.timedelta:
        """
        Retrieve the idle time of this pool member
        """
        return datetime.datetime.now().astimezone() - self.last_used

    def get_id(self) -> TIntPoolID:
        """
        Returns the ID of the pool member that will be used to lock the pool member. This ensures that no operations will
        overlap while this pool member is being created, modified, or cleaned.
        """
        return self.id

    async def request_shutdown(self) -> None:
        """
        Close the pool member.

        This should eventually cause a call to closed
        """
        self.shutting_down = True

    async def set_shutdown(self) -> None:
        """
        This pool member is now closed
        """
        self.shut_down = True
        self.shutting_down = True
        for listener in self.termination_listeners:
            await listener(self)


TPoolMember = TypeVar("TPoolMember", bound=PoolMember)


class PoolManager(abc.ABC, Generic[TPoolID, TIntPoolID, TPoolMember]):
    """
    A pool of slow objects

    Items are stored by id.

    We have internal id, that is used internally (storage, cleanup)
    We have pool id, that is used to request new instances
    Most often these are the same

    When a pool member is shutting down, it is considered to be invalid and we allow it to be replaced in the pool
    As such, the pool id does not necessarily uniquely identify an object.

    When the member signals it is down, we drop it from the pool.

    The natural ID  (member.get_id()) of a member is assumed to be the internal id of the member
    """

    def __init__(self) -> None:
        self.shut_down = False
        self.shutting_down = False

        self._locks: inmanta.util.NamedLock = inmanta.util.NamedLock()
        self.pool: dict[TIntPoolID, TPoolMember] = {}

    @property
    def running(self) -> bool:
        return not self.shutting_down and not self.shut_down

    @abc.abstractmethod
    def _id_to_internal(self, ext_id: TPoolID) -> TIntPoolID:
        """Convert an external id to an internal id"""
        pass

    def my_name(self) -> str:
        """Method to improve logging output by naming this executor"""
        return "PoolManager"

    def render_id(self, member: TPoolID) -> str:
        """Method to improve logging output by naming external ids"""
        # This method is not abstract to allow this class to be integrated with minimal effort
        # The default is not very informative, to force implementors to eventually override it
        return "PoolMember"

    def member_name(self, member: TPoolMember) -> str:
        """Method to improve logging output by naming the members, best kept consistent with render_id"""
        return self.render_id(member.get_id())

    def get_lock_name_for(self, member_id: TIntPoolID) -> str:
        """Convert the id into a string to obtain a lock"""
        return str(member_id)

    async def start(self) -> None:
        """
        Start the Pool Manager
        """
        pass

    async def request_shutdown(self) -> None:
        """
        Stop the Pool Manager.

        This implies nothing about the children
        """
        self.shut_down = True
        self.shutting_down = True

    async def join(self) -> None:
        """
        Wait for shutdown to be completed
        """
        pass

    async def notify_member_shutdown(self, pool_member: TPoolMember) -> bool:
        """
        a child has notified us that it is completely closed

        We take no lock, as we are either
        - in a crash situation and lock would not help
        - have gone through closing state so we are properly handled

        :return: if we effectively removed it from the pool
        """
        theid = pool_member.get_id()
        registered_for_id = self.pool.get(theid)
        if registered_for_id is None:
            # already dropped from cache
            return False
        elif registered_for_id != pool_member:
            # We have a stale instance that refused to close before
            # it was replaced and has now gone down
            return False
        else:
            self.pool.pop(theid)
            return True

    async def get(self, member_id: TPoolID) -> TPoolMember:
        """
        Returns a valid pool member for the given id
        """
        LOGGER.log(LOG_LEVEL_TRACE, "%s: requesting %s", self.my_name(), self.render_id(member_id))
        internal_id = self._id_to_internal(member_id)
        # Acquire a lock based on the executor's pool id
        async with self._locks.get(self.get_lock_name_for(internal_id)):
            it = self.pool.get(internal_id, None)
            if it is not None:
                if not it.shutting_down:
                    LOGGER.debug("%s: found existing %s", self.my_name(), self.render_id(member_id))
                    it.touch()
                    return it
                else:
                    await self.pre_replace(it)
            LOGGER.debug("%s: creating %s", self.my_name(), self.render_id(member_id))
            out = await self._create_or_replace(member_id, internal_id)
            LOGGER.debug("%s: created %s", self.my_name(), self.render_id(member_id))
            return out

    async def _create_or_replace(self, member_id: TPoolID, internal_id: TIntPoolID) -> TPoolMember:
        """
        MUST BE CALLED UNDER LOCK!
        """
        await self.pre_create_capacity_check(member_id)

        my_executor = await self.create_member(member_id)
        assert my_executor.id == internal_id

        self.pool[internal_id] = my_executor
        my_executor.termination_listeners.append(self.notify_member_shutdown)

        return my_executor

    async def pre_replace(self, member: PoolMember) -> None:
        """Hook method to join/handle items being replaced, called under lock"""
        pass

    async def pre_create_capacity_check(self, member_id: TPoolID) -> None:
        """
        Check if any members should be closed for the given id

        Will be called onder lock, is expected to call request_close
        """
        pass

    @abc.abstractmethod
    async def create_member(self, executor_id: TPoolID) -> TPoolMember:
        """Produce a fresh member for the given id"""
        pass


class SingleIdPoolManager(PoolManager[TPoolID, TPoolID, TPoolMember]):
    """Pool where internal and external id are the same"""

    def _id_to_internal(self, ext_id: TPoolID) -> TPoolID:
        return ext_id


class TimeBasedPoolManager(PoolManager[TPoolID, TIntPoolID, TPoolMember]):
    """
    Pool that will discard items after a specific period of not being used.

    Getting a member from the pool is considered use
    """

    def __init__(self, retention_time: int) -> None:
        super().__init__()

        # We keep a reference to the periodic cleanup task to prevent it
        # from disappearing mid-execution https://docs.python.org/3.11/library/asyncio-task.html#creating-tasks
        self.cleanup_job: Optional[asyncio.Task[None]] = None
        self.retention_time: int = retention_time
        # We keep track of the sleep function to be able to cancel it on shutdown
        # Without risking interrupting the cleanup itself
        self.shutdown_sleep: Optional[Task[None]] = None

    async def start(self) -> None:
        """
        Start the cleaning job of the Pool Manager
        """
        await super().start()
        self.cleanup_job = asyncio.create_task(self.cleanup_inactive_pool_members_task())

    async def request_shutdown(self) -> None:
        await super().request_shutdown()
        if self.shutdown_sleep is not None and not self.shutdown_sleep.done():
            self.shutdown_sleep.cancel("Shutting down cleanup task")

    async def join(self) -> None:
        """
        Wait for the cleaning job to terminate.
        """
        if self.cleanup_job is not None:
            await self.cleanup_job
        await super().join()

    async def cleanup_inactive_pool_members_task(self) -> None:
        """
        This task periodically cleans up idle pool member

        We split up `cleanup_inactive_pool_members` and `cleanup_inactive_pool_members_task` in order to be able to call the
        cleanup method in the test without being blocked in a loop.
        """
        try:
            while self.running:
                try:
                    sleep_interval = await self.cleanup_inactive_pool_members()
                except Exception:
                    # This should not happen, as the cleanup_inactive_pool_members should handle all exceptions
                    logging.exception("Unexpected error while cleaning up pool members")
                if self.running:
                    LOGGER.log(LOG_LEVEL_TRACE, "Manager will clean in %.2f seconds", sleep_interval)
                    # Allow wait to be cancelled on shutdown
                    self.shutdown_sleep = asyncio.create_task(asyncio.sleep(sleep_interval))
                    await self.shutdown_sleep
        except CancelledError:
            # We are woken up by shutdown
            pass

    async def cleanup_inactive_pool_members(self) -> float:
        """
        Cleans up idle pool member

        :return: When to run the cleaning next time: default retention time or lowest expiration time of one of the pool members
        """
        cleanup_start = datetime.datetime.now().astimezone()
        # Clean up anything older than this
        expiry = datetime.timedelta(seconds=self.retention_time)
        oldest_time = cleanup_start - expiry
        # Number of seconds relative to cleanup_start
        run_next_cleanup_job_in: float = float(self.retention_time)
        pool_members = list(self.pool.values())
        for pool_member in pool_members:
            try:
                if pool_member.can_be_cleaned_up() and pool_member.last_used < oldest_time and pool_member.running:
                    async with self._locks.get(self.get_lock_name_for(pool_member.get_id())):
                        # Check that the executor can still be cleaned up by the time we have acquired the lock
                        if pool_member.can_be_cleaned_up() and pool_member.last_used < oldest_time and pool_member.running:
                            LOGGER.debug(
                                "%s will be shutdown because it was inactive for %.2f, which is more than %d",
                                self.member_name(pool_member),
                                (cleanup_start - pool_member.last_used).total_seconds(),
                                self.retention_time,
                            )
                            await pool_member.request_shutdown()
                else:
                    # If this pool member expires sooner than the cleanup interval, schedule the next cleanup on that
                    # timestamp.
                    run_next_cleanup_job_in = min(
                        run_next_cleanup_job_in,
                        (pool_member.last_used - oldest_time).total_seconds(),
                    )
            except Exception:
                LOGGER.exception(
                    "An error occurred while cleaning the pool member `%s`",
                    pool_member.get_id(),
                )

        cleanup_end = datetime.datetime.now().astimezone()

        minimal_waiting_time: float = 0.5  # enforce minimal wait time to prevent bussy polling when can_be_cleaned_up == False
        return max(minimal_waiting_time, run_next_cleanup_job_in - (cleanup_end - cleanup_start).total_seconds())



"""
Copyright 2023 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    schema = """
    CREATE TYPE auth_method AS ENUM('database', 'oidc');

    CREATE TABLE IF NOT EXISTS public.user(
        id uuid PRIMARY KEY,
        username VARCHAR NOT NULL UNIQUE ,
        password_hash VARCHAR NOT NULL,
        auth_method auth_method NOT NULL
    );
    """

    await connection.execute(schema)



"""
Copyright 2020 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import logging
from typing import Optional

import docstring_parser

LOGGER = logging.getLogger(__name__)


class DocString:
    def __init__(self, doc_string: docstring_parser.Docstring) -> None:
        # Ignore attribute definitions with a syntax error
        new_meta = []
        for attr in doc_string.meta:
            # attr.args contains the part of an attribute definition in the docstring between the
            # colons splitted on white spaces. This check enforces the format :attr <attribute-name>:
            if len(attr.args) != 2 or attr.args[0] != "attr":
                LOGGER.warning("Failed to parse attribute: ':%s: %s'", " ".join(attr.args), attr.description)
            else:
                new_meta.append(attr)
        doc_string.meta = new_meta

        self._attr_description_map: dict[str, str] = {attr.args[1]: attr.description for attr in doc_string.meta}
        self._doc_string: docstring_parser.Docstring = doc_string

    def get_description(self) -> Optional[str]:
        """
        Return the general description in the docstring.
        """
        if self._doc_string.short_description is None:
            return None
        if self._doc_string.long_description is not None:
            return f"{self._doc_string.short_description}\n\n{self._doc_string.long_description}"
        else:
            return self._doc_string.short_description

    def get_description_for_attribute(self, attr_name: str) -> Optional[str]:
        """
        Return the description for a certain attribute.
        """
        return self._attr_description_map.get(attr_name, None)

    def get_attribute_description_map(self) -> dict[str, str]:
        """
        Return the dict which maps the attribute name to its description.
        """
        return dict(self._attr_description_map)


def parse_docstring(doc_string: str) -> DocString:
    """
    Parse the docstring of an entity.
    """
    doc_string = docstring_parser.parse(doc_string, style=docstring_parser.DocstringStyle.REST)
    return DocString(doc_string)
